"""Module defining the `QuantumNetwork` class and adjacent functions."""

import re
from functools import reduce
from operator import mul
from typing import List, Optional, Tuple, Union

import numpy as np

from qibo.backends import _check_backend
from qibo.config import raise_error


class QuantumNetwork:
    """This class stores the Choi operator of the quantum network as a tensor,
    which is an unique representation of the quantum network.

    A minimum quantum network is a quantum channel, which is a quantum network of the form
    :math:`J[n \\to m]`, where :math:`n` is the dimension of the input system ,
    and :math:`m` is the dimension of the output system.
    A quantum state is a quantum network of the form :math:`J[1 \\to n]`,
    such that the input system is trivial.
    An observable is a quantum network of the form :math:`J[n \\to 1]`,
    such that the output system is trivial.

    A quantum network may contain multiple input and output systems.
    For example, a "quantum comb" is a quantum network of the form :math:`J[n', n \\to m, m']`,
    which convert a quantum channel of the form :math:`J[n \\to m]`
    to a quantum channel of the form :math:`J[n' \\to m']`.

    Args:
        matrix (ndarray): input Choi operator.
        partition (List[int] or Tuple[int]): partition of ``matrix``.
        system_output (List[bool] or Tuple[bool], optional): mask on the output system of the
            Choi operator. If ``None``, defaults to
            ``(False,True,False,True,...)``, where ``len(system_output)=len(partition)``.
            Defaults to ``None``.
        pure (bool, optional): ``True`` when ``matrix`` is a "pure" representation (e.g. a pure
            state, a unitary operator, etc.), ``False`` otherwise. Defaults to ``False``.
        backend (:class:`qibo.backends.abstract.Backend`, optional): Backend to be used in
            calculations. If ``None``, defaults to :class:`qibo.backends.GlobalBackend`.
            Defaults to ``None``.
    """

    def __init__(
        self,
        matrix,
        partition: Union[List[int], Tuple[int]],
        system_output: Optional[Union[List[bool], Tuple[bool]]] = None,
        pure: bool = False,
        backend=None,
    ):
        self._run_checks(partition, system_output, pure)

        self._matrix = matrix
        self.partition = partition
        self.system_output = system_output
        self._pure = pure
        self._backend = backend
        self.dims = reduce(mul, self.partition)

        self._set_tensor_and_parameters()

    def matrix(self, backend=None):
        """Returns the Choi operator of the quantum network in matrix form.

        Args:
            backend (:class:`qibo.backends.abstract.Backend`, optional): Backend to be used
                to return the Choi operator. If ``None``, defaults to the backend defined
                when initializing the :class:`qibo.quantum_info.quantum_networks.QuantumNetwork`
                object. Defaults to ``None``.

        Returns:
            ndarray: Choi matrix of the quantum network.
        """
        if backend is None:  # pragma: no cover
            backend = self._backend

        return backend.cast(self._matrix, dtype=self._matrix.dtype)

    def is_pure(self):
        """Returns bool indicading if the Choi operator of the network is pure."""
        return self._pure

    def is_hermitian(
        self, order: Optional[Union[int, str]] = None, precision_tol: float = 1e-8
    ):
        """Returns bool indicating if the Choi operator :math:`\\mathcal{E}` of the network is Hermitian.

        Hermicity is calculated as distance between :math:`\\mathcal{E}` and
        :math:`\\mathcal{E}^{\\dagger}` with respect to a given norm.
        Default is the ``Hilbert-Schmidt`` norm (also known as ``Frobenius`` norm).

        For specifications on the other  possible values of the
        parameter ``order`` for the ``tensorflow`` backend, please refer to
        `tensorflow.norm <https://www.tensorflow.org/api_docs/python/tf/norm>`_.
        For all other backends, please refer to
        `numpy.linalg.norm <https://numpy.org/doc/stable/reference/generated/numpy.linalg.norm.html>`_.

        Args:
            order (str or int, optional): order of the norm. Defaults to ``None``.
            precision_tol (float, optional): threshold :math:`\\epsilon` that defines if
                Choi operator of the network is :math:`\\epsilon`-close to Hermicity in
                the norm given by ``order``. Defaults to :math:`10^{-8}`.

        Returns:
            bool: Hermiticity condition.
        """
        if precision_tol < 0.0:
            raise_error(
                ValueError,
                f"``precision_tol`` must be non-negative float, but it is {precision_tol}",
            )

        if order is None and self._backend.__class__.__name__ == "TensorflowBackend":
            order = "euclidean"

        self._matrix = self._full()
        self._pure = False

        reshaped = self._backend.cast(
            np.reshape(self._matrix, (self.dims, self.dims)), dtype=self._matrix.dtype
        )
        reshaped = self._backend.cast(
            np.transpose(np.conj(reshaped)) - reshaped, dtype=reshaped.dtype
        )
        norm = self._backend.calculate_norm_density_matrix(reshaped, order=order)

        return float(norm) <= precision_tol

    def is_unital(
        self, order: Optional[Union[int, str]] = None, precision_tol: float = 1e-8
    ):
        """Returns bool indicating if the Choi operator :math:`\\mathcal{E}` of the network is unital.

        Unitality is calculated as distance between the partial trace of :math:`\\mathcal{E}`
        and the Identity operator :math:`I`, with respect to a given norm.
        Default is the ``Hilbert-Schmidt`` norm (also known as ``Frobenius`` norm).

        For specifications on the other  possible values of the
        parameter ``order`` for the ``tensorflow`` backend, please refer to
        `tensorflow.norm <https://www.tensorflow.org/api_docs/python/tf/norm>`_.
        For all other backends, please refer to
        `numpy.linalg.norm <https://numpy.org/doc/stable/reference/generated/numpy.linalg.norm.html>`_.

        Args:
            order (str or int, optional): order of the norm. Defaults to ``None``.
            precision_tol (float, optional): threshold :math:`\\epsilon` that defines
                if Choi operator of the network is :math:`\\epsilon`-close to unitality
                in the norm given by ``order``. Defaults to :math:`10^{-8}`.

        Returns:
            bool: Unitality condition.
        """
        if precision_tol < 0.0:
            raise_error(
                ValueError,
                f"``precision_tol`` must be non-negative float, but it is {precision_tol}",
            )

        if order is None and self._backend.__class__.__name__ == "TensorflowBackend":
            order = "euclidean"

        self._matrix = self._full()
        self._pure = False

        partial_trace = self._einsum("jkjl -> kl", self._matrix)
        identity = self._backend.cast(
            np.eye(partial_trace.shape[0]), dtype=partial_trace.dtype
        )

        norm = self._backend.calculate_norm_density_matrix(
            partial_trace - identity,
            order=order,
        )

        return float(norm) <= precision_tol

    def is_causal(
        self, order: Optional[Union[int, str]] = None, precision_tol: float = 1e-8
    ):
        """Returns bool indicating if the Choi operator :math:`\\mathcal{E}` of the network satisfies the causal order condition.

        Causality is calculated as distance between partial trace of :math:`\\mathcal{E}`
        and the Identity operator :math:`I`, with respect to a given norm.
        Default is the ``Hilbert-Schmidt`` norm (also known as ``Frobenius`` norm).

        For specifications on the other  possible values of the
        parameter ``order`` for the ``tensorflow`` backend, please refer to
        `tensorflow.norm <https://www.tensorflow.org/api_docs/python/tf/norm>`_.
        For all other backends, please refer to
        `numpy.linalg.norm <https://numpy.org/doc/stable/reference/generated/numpy.linalg.norm.html>`_.

        Args:
            order (str or int, optional): order of the norm. Defaults to ``None``.
            precision_tol (float, optional): threshold :math:`\\epsilon` that defines
                if Choi operator of the network is :math:`\\epsilon`-close to causality
                in the norm given by ``order``. Defaults to :math:`10^{-8}`.

        Returns:
            bool: Causal order condition.
        """
        if precision_tol < 0.0:
            raise_error(
                ValueError,
                f"``precision_tol`` must be non-negative float, but it is {precision_tol}",
            )

        if order is None and self._backend.__class__.__name__ == "TensorflowBackend":
            order = "euclidean"

        self._matrix = self._full()
        self._pure = False

        partial_trace = self._einsum("jklk -> jl", self._matrix)
        identity = self._backend.cast(
            np.eye(partial_trace.shape[0]), dtype=partial_trace.dtype
        )

        norm = self._backend.calculate_norm_density_matrix(
            partial_trace - identity,
            order=order,
        )

        return float(norm) <= precision_tol

    def is_positive_semidefinite(self, precision_tol: float = 1e-8):
        """Returns bool indicating if Choi operator :math:`\\mathcal{E}` of the network is positive-semidefinite.

        Args:
            precision_tol (float, optional): threshold value used to check if eigenvalues of
                the Choi operator :math:`\\mathcal{E}` are such that
                :math:`\\textup{eigenvalues}(\\mathcal{E}) >= - \\textup{precision_tol}`.
                Note that this parameter can be set to negative values.
                Defaults to :math:`0.0`.

        Returns:
            bool: Positive-semidefinite condition.
        """
        self._matrix = self._full()
        self._pure = False

        reshaped = np.reshape(self._matrix, (self.dims, self.dims))

        if self.is_hermitian():
            eigenvalues = np.linalg.eigvalsh(reshaped)
        else:
            if self._backend.__class__.__name__ in [
                "CupyBackend",
                "CuQuantumBackend",
            ]:  # pragma: no cover
                reshaped = np.array(reshaped.tolist(), dtype=reshaped.dtype)
            eigenvalues = np.linalg.eigvals(reshaped)

        return all(eigenvalue >= -precision_tol for eigenvalue in eigenvalues)

    def is_channel(
        self,
        order: Optional[Union[int, str]] = None,
        precision_tol_causal: float = 1e-8,
        precision_tol_psd: float = 1e-8,
    ):
        """Returns bool indicating if Choi operator :math:`\\mathcal{E}` is a channel.

        Args:
            order (int or str, optional): order of the norm used to calculate causality.
                Defaults to ``None``.
            precision_tol_causal (float, optional): threshold :math:`\\epsilon` that defines if
                Choi operator of the network is :math:`\\epsilon`-close to causality in the norm
                given by ``order``. Defaults to :math:`10^{-8}`.
            precision_tol_psd (float, optional): threshold value used to check if eigenvalues of
                the Choi operator :math:`\\mathcal{E}` are such that
                :math:`\\textup{eigenvalues}(\\mathcal{E}) >= \\textup{precision_tol_psd}`.
                Note that this parameter can be set to negative values.
                Defaults to :math:`0.0`.

        Returns:
            bool: Channel condition.
        """
        return self.is_causal(
            order, precision_tol_causal
        ) and self.is_positive_semidefinite(precision_tol_psd)

    def apply(self, state):
        """Apply the Choi operator :math:`\\mathcal{E}` to ``state`` :math:`\\varrho`.

        It is assumed that ``state`` :math:`\\varrho` is a density matrix.

        Args:
            state (ndarray): density matrix of a ``state``.

        Returns:
            ndarray: Resulting state :math:`\\mathcal{E}(\\varrho)`.
        """
        matrix = self._backend.cast(self._matrix, copy=True)

        if self.is_pure():
            return self._einsum("kj,ml,jl -> km", matrix, np.conj(matrix), state)

        return self._einsum("jklm,km -> jl", matrix, state)

    def link_product(self, second_network, subscripts: str = "ij,jk -> ik"):
        """Link product between two quantum networks.

        The link product is not commutative. Here, we assume that
        :math:`A.\\textup{link_product}(B)` means "applying :math:`B` to :math:`A`".
        However, the ``link_product`` is associative, so we override the `@` operation
        in order to simplify notation.

        Args:
            second_network (:class:`qibo.quantum_info.quantum_networks.QuantumNetwork`): Quantum
                network to be applied to the original network.
            subscripts (str, optional): Specifies the subscript for summation using
                the Einstein summation convention. For more details, please refer to
                `numpy.einsum <https://numpy.org/doc/stable/reference/generated/numpy.einsum.html>`_.

        Returns:
            :class:`qibo.quantum_info.quantum_networks.QuantumNetwork`: Quantum network resulting
                from the link product between two quantum networks.
        """
        if not isinstance(second_network, QuantumNetwork):
            raise_error(
                TypeError,
                "It is not possible to implement link product of a "
                + "``QuantumNetwork`` with a non-``QuantumNetwork``.",
            )

        if not isinstance(subscripts, str):
            raise_error(
                TypeError,
                f"subscripts must be type str, but it is type {type(subscripts)}.",
            )

        subscripts = subscripts.replace(" ", "")
        pattern_two, pattern_four = self._check_subscript_pattern(subscripts)
        channel_subscripts = pattern_two and subscripts[1] == subscripts[3]
        inv_subscripts = pattern_two and subscripts[0] == subscripts[4]
        super_subscripts = (
            pattern_four
            and subscripts[1] == subscripts[5]
            and subscripts[2] == subscripts[6]
        )

        if not channel_subscripts and not inv_subscripts and not super_subscripts:
            raise_error(
                NotImplementedError,
                "Subscripts do not match any implemented pattern.",
            )

        first_matrix = self._full()
        second_matrix = second_network._full()  # pylint: disable=W0212

        if super_subscripts:
            cexpr = "jklmnopq,klop->jmnq"
            return QuantumNetwork(
                self._einsum(cexpr, first_matrix, second_matrix),
                [self.partition[0] + self.partition[-1]],
            )

        cexpr = "jkab,klbc->jlac"

        if inv_subscripts:
            return QuantumNetwork(
                self._einsum(cexpr, second_matrix, first_matrix),
                [second_network.partition[0], self.partition[1]],
            )

        return QuantumNetwork(
            self._einsum(cexpr, first_matrix, second_matrix),
            [self.partition[0], second_network.partition[1]],
        )

    def copy(self):
        """Returns a copy of the :class:`qibo.quantum_info.quantum_networks.QuantumNetwork` object."""
        return self.__class__(
            np.copy(self._matrix),
            partition=self.partition,
            system_output=self.system_output,
            pure=self._pure,
            backend=self._backend,
        )

    def to_full(self, backend=None):
        """Convert the internal representation to the full Choi operator of the network.

        Returns:
            (:class:`qibo.quantum_info.quantum_networks.QuantumNetwork`): The full representation
            of the Quantum network.
        """
        if backend is None:  # pragma: no cover
            backend = self._backend

        if self.is_pure():
            self._matrix = self._full()
            self._pure = False

        return self.matrix(backend)

    def __add__(self, second_network):
        """Add two Quantum Networks by adding their Choi operators.

        This operation always returns a non-pure Quantum Network.

        Args:
            second_network (:class:`qibo.quantum_info.quantum_networks.QuantumNetwork`): Quantum
                network to be added to the original network.

        Returns:
            (:class:`qibo.quantum_info.quantum_networks.QuantumNetwork`): Quantum network resulting
                from the summation of two Choi operators.
        """
        if not isinstance(second_network, QuantumNetwork):
            raise_error(
                TypeError,
                "It is not possible to add a object of type ``QuantumNetwork`` "
                + f"and and object of type ``{type(second_network)}``.",
            )

        if self._full().shape != second_network._full().shape:
            raise_error(
                ValueError,
                f"The Choi operators must have the same shape, but {self._matrix.shape} != "
                + f"{second_network.matrix(second_network._backend).shape}.",
            )

        if self.system_output != second_network.system_output:
            raise_error(ValueError, "The networks must have the same output system.")

        new_first_matrix = self._full()
        new_second_matrix = second_network._full()

        return QuantumNetwork(
            new_first_matrix + new_second_matrix,
            self.partition,
            self.system_output,
            pure=False,
            backend=self._backend,
        )

    def __mul__(self, number: Union[float, int]):
        """Returns quantum network with its Choi operator multiplied by a scalar.

        If the quantum network is pure and ``number > 0.0``, the method returns a pure quantum
        network with its Choi operator multiplied by the square root of ``number``.
        This is equivalent to multiplying `self.to_full()` by the ``number``.
        Otherwise, this method will return a full quantum network.

        Args:
            number (float or int): scalar to multiply the Choi operator of the network with.

        Returns:
            :class:`qibo.quantum_info.quantum_networks.QuantumNetwork`: Quantum network with its
                Choi operator multiplied by ``number``.
        """
        if not isinstance(number, (float, int)):
            raise_error(
                TypeError,
                "It is not possible to multiply a ``QuantumNetwork`` by a non-scalar.",
            )

        if self.is_pure() and number > 0.0:
            return QuantumNetwork(
                np.sqrt(number) * self.matrix(backend=self._backend),
                partition=self.partition,
                system_output=self.system_output,
                pure=True,
                backend=self._backend,
            )

        matrix = self._full()

        return QuantumNetwork(
            number * matrix,
            partition=self.partition,
            system_output=self.system_output,
            pure=False,
            backend=self._backend,
        )

    def __rmul__(self, number: Union[float, int]):
        """"""
        return self.__mul__(number)

    def __truediv__(self, number: Union[float, int]):
        """Returns quantum network with its Choi operator divided by a scalar.

        If the quantum network is pure and ``number > 0.0``, the method returns a pure quantum
        network with its Choi operator divided by the square root of ``number``.
        This is equivalent to dividing `self.to_full()` by the ``number``.
        Otherwise, this method will return a full quantum network.

        Args:
            number (float or int): scalar to divide the Choi operator of the network with.

        Returns:
            :class:`qibo.quantum_info.quantum_networks.QuantumNetwork`: Quantum network with its
                Choi operator divided by ``number``.
        """
        if not isinstance(number, (float, int)):
            raise_error(
                TypeError,
                "It is not possible to divide a ``QuantumNetwork`` by a non-scalar.",
            )

        number = np.sqrt(number) if self.is_pure() and number > 0.0 else number

        return QuantumNetwork(
            self.matrix(backend=self._backend) / number,
            partition=self.partition,
            system_output=self.system_output,
            pure=self.is_pure(),
            backend=self._backend,
        )

    def __matmul__(self, second_network):
        """Defines matrix multiplication between two ``QuantumNetwork`` objects.

        If ``self.partition == second_network.partition in [2, 4]``, this method is overwritten by
        :meth:`qibo.quantum_info.quantum_networks.QuantumNetwork.link_product`.

        Args:
            second_network (:class:`qibo.quantum_info.quantum_networks.QuantumNetwork`):

        Returns:
            :class:`qibo.quantum_info.quantum_networks.QuantumNetwork`: Quantum network resulting
                from the link product operation.
        """
        if not isinstance(second_network, QuantumNetwork):
            raise_error(
                TypeError,
                "It is not possible to implement matrix multiplication of a "
                + "``QuantumNetwork`` by a non-``QuantumNetwork``.",
            )

        if len(self.partition) == 2:  # `self` is a channel
            if len(second_network.partition) != 2:
                raise_error(
                    ValueError,
                    f"`QuantumNetwork {second_network} is assumed to be a channel, but it is not. "
                    + "Use `link_product` method to specify the subscript.",
                )
            if self.partition[1] != second_network.partition[0]:
                raise_error(
                    ValueError,
                    "partitions of the networks do not match: "
                    + f"{self.partition[1]} != {second_network.partition[0]}.",
                )

            subscripts = "jk,kl -> jl"

        elif len(self.partition) == 4:  # `self` is a super-channel
            if len(second_network.partition) != 2:
                raise_error(
                    ValueError,
                    f"`QuantumNetwork {second_network} is assumed to be a channel, but it is not. "
                    + "Use `link_product` method to specify the subscript.",
                )
            if self.partition[1] != second_network.partition[0]:
                raise_error(
                    ValueError,
                    "Systems of the channel do not match the super-channel: "
                    + f"{self.partition[1], self.partition[2]} != "
                    + f"{second_network.partition[0],second_network.partition[1]}.",
                )

            subscripts = "jklm,kl -> jm"
        else:
            raise_error(
                NotImplementedError,
                "`partitions` do not match any implemented pattern``. "
                + "Use `link_product` method to specify the subscript.",
            )

        return self.link_product(second_network, subscripts=subscripts)

    def __str__(self):
        """Method to define how to print relevant information of the quantum network."""
        string_in = ", ".join(
            [
                str(self.partition[k])
                for k in range(len(self.partition))
                if not self.system_output[k]
            ]
        )

        string_out = ", ".join(
            [
                str(self.partition[k])
                for k in range(len(self.partition))
                if self.system_output[k]
            ]
        )

        return f"J[{string_in} -> {string_out}]"

    def _run_checks(self, partition, system_output, pure):
        """Checks if all inputs are correct in type and value."""
        if not isinstance(partition, (list, tuple)):
            raise_error(
                TypeError,
                "``partition`` must be type ``tuple`` or ``list``, "
                + f"but it is type ``{type(partition)}``.",
            )

        if any(not isinstance(party, int) for party in partition):
            raise_error(
                ValueError,
                "``partition`` must be a ``tuple`` or ``list`` of positive integers, "
                + "but contains non-integers.",
            )

        if any(party <= 0 for party in partition):
            raise_error(
                ValueError,
                "``partition`` must be a ``tuple`` or ``list`` of positive integers, "
                + "but contains non-positive integers.",
            )

        if system_output is not None and len(system_output) != len(partition):
            raise_error(
                ValueError,
                "``len(system_output)`` must be the same as ``len(partition)``, "
                + f"but {len(system_output)} != {len(partition)}.",
            )

        if not isinstance(pure, bool):
            raise_error(
                TypeError,
                f"``pure`` must be type ``bool``, but it is type ``{type(pure)}``.",
            )

    def _set_tensor_and_parameters(self):
        """Sets tensor based on inputs."""
        self._backend = _check_backend(self._backend)

        self._einsum = self._backend.np.einsum

        if isinstance(self.partition, list):
            self.partition = tuple(self.partition)

        try:
            if self._pure:
                self._matrix = np.reshape(self._matrix, self.partition)
            else:
                matrix_partition = self.partition * 2
                self._matrix = np.reshape(self._matrix, matrix_partition)
        except:
            raise_error(
                ValueError,
                "``partition`` does not match the shape of the input matrix. "
                + f"Cannot reshape matrix of size {self._matrix.shape} to partition {self.partition}",
            )

        if self.system_output is None:
            self.system_output = [
                True,
            ] * len(self.partition)
            for k in range(len(self.partition) // 2):
                self.system_output[k * 2] = False
            self.system_output = tuple(self.system_output)
        else:
            self.system_output = tuple(self.system_output)

    def _full(self):
        """Reshapes input matrix based on purity."""
        matrix = self._backend.cast(self._matrix, copy=True)

        if self.is_pure():
            matrix = self._einsum("jk,lm -> kjml", matrix, np.conj(matrix))

        return matrix

    def _check_subscript_pattern(self, subscripts: str):
        """Checks if input subscript match any implemented pattern."""
        braket = "[a-z]"
        pattern_two = re.compile(braket * 2 + "," + braket * 2 + "->" + braket * 2)
        pattern_four = re.compile(braket * 4 + "," + braket * 2 + "->" + braket * 2)

        return bool(re.match(pattern_two, subscripts)), bool(
            re.match(pattern_four, subscripts)
        )



import math

from qibo import gates
from qibo.config import raise_error
from qibo.models.circuit import Circuit


def QFT(nqubits, with_swaps=True, accelerators=None) -> Circuit:
    """Creates a circuit that implements the Quantum Fourier Transform.

    Args:
        nqubits (int): Number of qubits in the circuit.
        with_swaps (bool): Use SWAP gates at the end of the circuit so that the
            qubit order in the final state is the same as the initial state.
        accelerators (dict): Accelerator device dictionary in order to use a
            distributed circuit
            If ``None`` a simple (non-distributed) circuit will be used.

    Returns:
        A qibo.models.Circuit that implements the Quantum Fourier Transform.

    Example:
        .. testcode::

            import numpy as np
            from qibo.models import QFT
            nqubits = 6
            c = QFT(nqubits)
            # Random normalized initial state vector
            init_state = np.random.random(2 ** nqubits) + 1j * np.random.random(2 ** nqubits)
            init_state = init_state / np.sqrt((np.abs(init_state)**2).sum())
            # Execute the circuit
            final_state = c(init_state)
    """
    if accelerators is not None:
        if not with_swaps:
            raise_error(
                NotImplementedError,
                "Distributed QFT is only implemented " "with SWAPs.",
            )
        return _DistributedQFT(nqubits, accelerators)

    circuit = Circuit(nqubits)
    for i1 in range(nqubits):
        circuit.add(gates.H(i1))
        for i2 in range(i1 + 1, nqubits):
            theta = math.pi / 2 ** (i2 - i1)
            circuit.add(gates.CU1(i2, i1, theta))

    if with_swaps:
        for i in range(nqubits // 2):
            circuit.add(gates.SWAP(i, nqubits - i - 1))

    return circuit


def _DistributedQFT(nqubits, accelerators=None):
    """QFT with the order of gates optimized for reduced multi-device communication."""
    circuit = Circuit(nqubits, accelerators)
    icrit = nqubits // 2 + nqubits % 2
    if accelerators is not None:
        circuit.global_qubits = range(circuit.nlocal, nqubits)  # pylint: disable=E1101
        if icrit < circuit.nglobal:  # pylint: disable=E1101
            raise_error(
                NotImplementedError,
                f"Cannot implement QFT for {nqubits} qubits "
                + f"using {circuit.nglobal} global qubits.",
            )  # pylint: disable=E1101

    for i1 in range(nqubits):
        if i1 < icrit:
            i1eff = i1
        else:
            i1eff = nqubits - i1 - 1
            circuit.add(gates.SWAP(i1, i1eff))

        circuit.add(gates.H(i1eff))
        for i2 in range(i1 + 1, nqubits):
            theta = math.pi / 2 ** (i2 - i1)
            circuit.add(gates.CU1(i2, i1eff, theta))

    return circuit



"""
Gates use ``einsum`` to apply gates to state vectors. The einsum string that
specifies the contraction indices is created using the following methods and
used by :meth:`qibo.backends.numpy.NumpyEngine.apply_gate`.
"""

from qibo.config import EINSUM_CHARS, raise_error


def prepare_strings(qubits, nqubits):
    if nqubits + len(qubits) > len(EINSUM_CHARS):  # pragma: no cover
        raise_error(NotImplementedError, "Not enough einsum characters.")

    inp = list(EINSUM_CHARS[:nqubits])
    out = inp[:]
    trans = list(EINSUM_CHARS[nqubits : nqubits + len(qubits)])
    for i, q in enumerate(qubits):
        trans.append(inp[q])
        out[q] = trans[i]

    inp = "".join(inp)
    out = "".join(out)
    trans = "".join(trans)
    rest = EINSUM_CHARS[nqubits + len(qubits) :]
    return inp, out, trans, rest


def apply_gate_string(qubits, nqubits):
    inp, out, trans, _ = prepare_strings(qubits, nqubits)
    return f"{inp},{trans}->{out}"


def apply_gate_density_matrix_string(qubits, nqubits):
    inp, out, trans, rest = prepare_strings(qubits, nqubits)
    if nqubits > len(rest):  # pragma: no cover
        raise_error(NotImplementedError, "Not enough einsum characters.")

    trest = rest[:nqubits]
    left = f"{inp}{trest},{trans}->{out}{trest}"
    right = f"{trest}{inp},{trans}->{trest}{out}"
    return left, right


def apply_gate_density_matrix_controlled_string(qubits, nqubits):
    inp, out, trans, rest = prepare_strings(qubits, nqubits)
    if nqubits > len(rest):  # pragma: no cover
        raise_error(NotImplementedError, "Not enough einsum characters.")

    trest, c = rest[:nqubits], rest[nqubits]
    left = f"{c}{inp}{trest},{trans}->{c}{out}{trest}"
    right = f"{c}{trest}{inp},{trans}->{c}{trest}{out}"
    return left, right


def control_order(gate, nqubits):
    loop_start = 0
    order = list(gate.control_qubits)
    targets = list(gate.target_qubits)
    for control in gate.control_qubits:
        for i in range(loop_start, control):
            order.append(i)
        loop_start = control + 1
        for i, t in enumerate(gate.target_qubits):
            if t > control:
                targets[i] -= 1
    for i in range(loop_start, nqubits):
        order.append(i)
    return order, targets


def control_order_density_matrix(gate, nqubits):
    ncontrol = len(gate.control_qubits)
    order, targets = control_order(gate, nqubits)
    additional_order = [x + len(order) for x in order]
    order_dm = (
        order[:ncontrol]
        + list(additional_order[:ncontrol])
        + order[ncontrol:]
        + list(additional_order[ncontrol:])
    )
    return order_dm, targets


def reverse_order(order):
    rorder = len(order) * [0]
    for i, r in enumerate(order):
        rorder[r] = i
    return rorder



"""Custom exceptions raised in transpiler routines."""


class BlockingError(Exception):
    """Raise when an error occurs in the blocking procedure"""


class ConnectivityError(Exception):
    """Raise for an error in the connectivity"""


class DecompositionError(Exception):
    """A decomposition error is raised when, during transpiling,
    gates are not correctly decomposed in native gates"""


class PlacementError(Exception):
    """Raise for an error in the initial qubit placement"""


class TranspilerPipelineError(Exception):
    """Raise when an error occurs in the transpiler pipeline"""



from abc import ABC, abstractmethod
from typing import Tuple

import networkx as nx

from qibo.models import Circuit


class Placer(ABC):
    @abstractmethod
    def __init__(self, connectivity: nx.Graph, *args):
        """A placer implements the initial logical-physical qubit mapping"""

    @abstractmethod
    def __call__(self, circuit: Circuit, *args) -> dict:
        """Find initial qubit mapping

        Args:
            circuit (:class:`qibo.models.circuit.Circuit`): circuit to be mapped.

        Returns:
            (dict): dictionary containing the initial logical to physical qubit mapping.
        """


class Router(ABC):
    @abstractmethod
    def __init__(self, connectivity: nx.Graph, *args):
        """A router implements the mapping of a circuit on a specific hardware."""

    @abstractmethod
    def __call__(
        self, circuit: Circuit, initial_layout: dict, *args
    ) -> Tuple[Circuit, dict]:
        """Match circuit to hardware connectivity.

        Args:
            circuit (qibo.models.Circuit): circuit to be routed.
            initial_layout (dict): dictionary containing the initial logical to physical qubit mapping.

        Returns:
            (:class:`qibo.models.circuit.Circuit`, dict): routed circuit and dictionary containing the final logical to physical qubit mapping.
        """


class Optimizer(ABC):
    """An optimizer tries to reduce the number of gates during transpilation."""

    @abstractmethod
    def __call__(self, circuit: Circuit, *args) -> Circuit:
        """Optimize transpiled circuit.

        Args:
            circuit (:class:`qibo.models.circuit.Circuit`): circuit to be optimized

        Returns:
            (:class:`qibo.models.circuit.Circuit`): circuit with optimized number of gates.
        """



"""Module with the most common linear algebra operations for quantum information."""

from qibo.config import raise_error


def commutator(operator_1, operator_2):
    """Returns the commutator of ``operator_1`` and ``operator_2``.

    The commutator of two matrices :math:`A` and :math:`B` is given by

    .. math::
        [A, B] = A \\, B - B \\, A \\,.

    Args:
        operator_1 (ndarray): First operator.
        operator_2 (ndarray): Second operator.

    Returns:
        ndarray: Commutator of ``operator_1`` and ``operator_2``.
    """
    if (
        (len(operator_1.shape) >= 3)
        or (len(operator_1) == 0)
        or (len(operator_1.shape) == 2 and operator_1.shape[0] != operator_1.shape[1])
    ):
        raise_error(
            TypeError,
            f"``operator_1`` must have shape (k,k), but have shape {operator_1.shape}.",
        )

    if (
        (len(operator_2.shape) >= 3)
        or (len(operator_2) == 0)
        or (len(operator_2.shape) == 2 and operator_2.shape[0] != operator_2.shape[1])
    ):
        raise_error(
            TypeError,
            f"``operator_2`` must have shape (k,k), but have shape {operator_2.shape}.",
        )

    if operator_1.shape != operator_2.shape:
        raise_error(
            TypeError,
            "``operator_1`` and ``operator_2`` must have the same shape, "
            + f"but {operator_1.shape} != {operator_2.shape}",
        )

    return operator_1 @ operator_2 - operator_2 @ operator_1


def anticommutator(operator_1, operator_2):
    """Returns the anticommutator of ``operator_1`` and ``operator_2``.

    The anticommutator of two matrices :math:`A` and :math:`B` is given by

    .. math::
        \\{A, B\\} = A \\, B + B \\, A \\,.

    Args:
        operator_1 (ndarray): First operator.
        operator_2 (ndarray): Second operator.

    Returns:
        ndarray: Anticommutator of ``operator_1`` and ``operator_2``.
    """
    if (
        (len(operator_1.shape) >= 3)
        or (len(operator_1) == 0)
        or (len(operator_1.shape) == 2 and operator_1.shape[0] != operator_1.shape[1])
    ):
        raise_error(
            TypeError,
            f"``operator_1`` must have shape (k,k), but have shape {operator_1.shape}.",
        )

    if (
        (len(operator_2.shape) >= 3)
        or (len(operator_2) == 0)
        or (len(operator_2.shape) == 2 and operator_2.shape[0] != operator_2.shape[1])
    ):
        raise_error(
            TypeError,
            f"``operator_2`` must have shape (k,k), but have shape {operator_2.shape}.",
        )

    if operator_1.shape != operator_2.shape:
        raise_error(
            TypeError,
            "``operator_1`` and ``operator_2`` must have the same shape, "
            + f"but {operator_1.shape} != {operator_2.shape}",
        )

    return operator_1 @ operator_2 + operator_2 @ operator_1



"""
Resources for parallel circuit evaluation.
"""

from typing import Iterable

from joblib import Parallel, delayed

from qibo.backends import _check_backend
from qibo.config import raise_error


def parallel_execution(circuit, states, processes=None, backend=None):
    """Execute circuit for multiple states.

    Example:
        .. code-block:: python

            import qibo
            qibo.set_backend('qibojit')
            from qibo import models, set_threads
            from qibo.parallel import parallel_execution
            import numpy as np
            # create circuit
            nqubits = 22
            circuit = models.QFT(nqubits)
            # create random states
            states = [ np.random.random(2**nqubits) for i in range(5)]
            # set threads to 1 per process (optional, requires tuning)
            set_threads(1)
            # execute in parallel
            results = parallel_execution(circuit, states, processes=2)

    Args:
        circuit (qibo.models.Circuit): the input circuit.
        states (list): list of states for the circuit evaluation.
        processes (int): number of processes for parallel evaluation.

    Returns:
        Circuit evaluation for input states.
    """
    backend = _check_backend(backend)

    if states is None or not isinstance(states, list):  # pragma: no cover
        raise_error(TypeError, "states must be a list.")

    def operation(state, circuit):
        backend.set_threads(backend.nthreads)
        return backend.execute_circuit(circuit, state)

    results = Parallel(n_jobs=processes, prefer="threads")(
        delayed(operation)(state, circuit) for state in states
    )

    return results


def parallel_circuits_execution(
    circuits, states=None, nshots=1000, processes=None, backend=None
):
    """Execute multiple circuits

    Example:
        .. code-block:: python

            import qibo
            qibo.set_backend('qibojit')
            from qibo import models, set_threads
            from qibo.parallel import parallel_circuits_execution
            import numpy as np
            # create different circuits
            circuits = [models.QFT(n) for n in range(5, 16)]
            # set threads to 1 per process (optional, requires tuning)
            set_threads(1)
            # execute in parallel
            results = parallel_circuits_execution(circuits, processes=2)

    Args:
        circuits (list): list of circuits to execute.
        states (optional, list): list of states to use as initial for each circuit.
            Must have the same length as ``circuits``.
            If one state is given it is used on all circuits.
            If not given the default initial state on all circuits.
        nshots (int): Number of shots when performing measurements, same for all circuits.
        processes (int): number of processes for parallel evaluation.

    Returns:
        Circuit evaluation for input states.
    """
    backend = _check_backend(backend)

    if not isinstance(circuits, Iterable):  # pragma: no cover
        raise_error(TypeError, "circuits must be iterable.")

    if (
        isinstance(states, (list, tuple))
        and isinstance(circuits, (list, tuple))
        and len(states) != len(circuits)
    ):
        raise_error(ValueError, "states must have the same length as circuits.")
    elif states is not None and not isinstance(states, Iterable):
        raise_error(TypeError, "states must be iterable.")

    def operation(circuit, state):
        backend.set_threads(backend.nthreads)
        return backend.execute_circuit(circuit, state, nshots)

    if states is None or isinstance(states, backend.tensor_types):
        results = Parallel(n_jobs=processes, prefer="threads")(
            delayed(operation)(circuit, states) for circuit in circuits
        )
    else:
        results = Parallel(n_jobs=processes, prefer="threads")(
            delayed(operation)(circuit, state)
            for circuit, state in zip(circuits, states)
        )

    return results


def parallel_parametrized_execution(
    circuit, parameters, initial_state=None, processes=None, backend=None
):
    """Execute circuit for multiple parameters and fixed initial_state.

    Example:
        .. code-block:: python

            import qibo
            qibo.set_backend('qibojit')
            from qibo import models, gates, set_threads
            from qibo.parallel import parallel_parametrized_execution
            import numpy as np
            # create circuit
            nqubits = 6
            nlayers = 2
            circuit = models.Circuit(nqubits)
            for l in range(nlayers):
                circuit.add((gates.RY(q, theta=0) for q in range(nqubits)))
                circuit.add((gates.CZ(q, q+1) for q in range(0, nqubits-1, 2)))
                circuit.add((gates.RY(q, theta=0) for q in range(nqubits)))
                circuit.add((gates.CZ(q, q+1) for q in range(1, nqubits-2, 2)))
                circuit.add(gates.CZ(0, nqubits-1))
            circuit.add((gates.RY(q, theta=0) for q in range(nqubits)))
            # create random parameters
            size = len(circuit.get_parameters())
            parameters = [ np.random.uniform(0, 2*np.pi, size) for _ in range(10) ]
            # set threads to 1 per process (optional, requires tuning)
            set_threads(1)
            # execute in parallel
            results = parallel_parametrized_execution(circuit, parameters, processes=2)

    Args:
        circuit (qibo.models.Circuit): the input circuit.
        parameters (list): list of parameters for the circuit evaluation.
        initial_state (np.array): initial state for the circuit evaluation.
        processes (int): number of processes for parallel evaluation.
            This corresponds to the number of threads, if a single thread is used
            for each circuit evaluation. If more threads are used for each circuit
            evaluation then some tuning may be required to obtain optimal performance.
            Default is ``None`` which corresponds to a single thread.

    Returns:
        Circuit evaluation for input parameters.
    """
    backend = _check_backend(backend)

    if not isinstance(parameters, list):  # pragma: no cover
        raise_error(TypeError, "parameters must be a list.")

    def operation(params, circuit, state):
        backend.set_threads(backend.nthreads)
        if state is not None:
            state = backend.cast(state, copy=True)
        circuit.set_parameters(params)
        return backend.execute_circuit(circuit, state)

    results = Parallel(n_jobs=processes, prefer="threads")(
        delayed(operation)(param, circuit.copy(deep=True), initial_state)
        for param in parameters
    )

    return results



from copy import deepcopy
from enum import Enum, auto
from typing import Optional

import numpy as np

from qibo.hamiltonians import Hamiltonian
from qibo.models.dbi.utils import *
from qibo.models.dbi.utils_scheduling import (
    grid_search_step,
    hyperopt_step,
    polynomial_step,
    simulated_annealing_step,
)


class DoubleBracketGeneratorType(Enum):
    """Define DBF evolution."""

    canonical = auto()
    """Use canonical commutator."""
    single_commutator = auto()
    """Use single commutator."""
    group_commutator = auto()
    """Use group commutator approximation"""
    group_commutator_third_order = auto()
    """Implements: $e^{\frac{\\sqrt{5}-1}{2}isH}e^{\frac{\\sqrt{5}-1}{2}isD}e^{-isH}e^{isD}e^{\frac{3-\\sqrt{5}}{2}isH}e^{isD}
    \approx e^{-s^2[H,D]} + O(s^4)$
    which is equation (8) in https://arxiv.org/abs/2111.12177]
    s must be taken as $\\sqrt{s}$ to approximate the flow using the commutator
    """


class DoubleBracketCostFunction(str, Enum):
    """Define the DBI cost function."""

    off_diagonal_norm = "off_diagonal_norm"
    """Use off-diagonal norm as cost function."""
    least_squares = "least_squares"
    """Use least squares as cost function."""
    energy_fluctuation = "energy_fluctuation"
    """Use energy fluctuation as cost function."""


class DoubleBracketScheduling(Enum):
    """Define the DBI scheduling strategies."""

    hyperopt = hyperopt_step
    """Use hyperopt package."""
    grid_search = grid_search_step
    """Use greedy grid search."""
    polynomial_approximation = polynomial_step
    """Use polynomial expansion (analytical) of the loss function."""
    simulated_annealing = simulated_annealing_step
    """Use simulated annealing algorithm"""


class DoubleBracketIteration:
    """
    Class implementing the Double Bracket iteration algorithm.
    For more details, see https://arxiv.org/pdf/2206.11772.pdf

    Args:
        hamiltonian (Hamiltonian): Starting Hamiltonian;
        mode (DoubleBracketGeneratorType): type of generator of the evolution.

    Example:
        .. testcode::

            from qibo.models.dbi.double_bracket import DoubleBracketIteration, DoubleBracketGeneratorType
            from qibo.quantum_info import random_hermitian
            from qibo.hamiltonians import Hamiltonian

            nqubits = 4
            h0 = random_hermitian(2**nqubits, seed=2)
            dbf = DoubleBracketIteration(Hamiltonian(nqubits=nqubits, matrix=h0))

            # diagonalized matrix
            dbf.h
    """

    def __init__(
        self,
        hamiltonian: Hamiltonian,
        mode: DoubleBracketGeneratorType = DoubleBracketGeneratorType.canonical,
        scheduling: DoubleBracketScheduling = DoubleBracketScheduling.grid_search,
        cost: DoubleBracketCostFunction = DoubleBracketCostFunction.off_diagonal_norm,
        ref_state: np.array = None,
    ):
        self.h = hamiltonian
        self.h0 = deepcopy(self.h)
        self.mode = mode
        self.scheduling = scheduling
        self.cost = cost
        self.ref_state = ref_state
        """
        Args:
            hamiltonian (Hamiltonian): Starting Hamiltonian;
            mode (DoubleBracketGeneratorType): type of generator of the evolution.
            scheduling (DoubleBracketScheduling): type of scheduling strategy.
            cost (DoubleBracketCost): type of cost function.
            ref_state (np.array): reference state for computing the energy fluctuation.
        """

    def __call__(
        self, step: float, mode: DoubleBracketGeneratorType = None, d: np.array = None
    ):
        r"""We use convention that $H' = U^\dagger H U$ where $U=e^{-sW}$ with $W=[D,H]$
        (or depending on `mode` an approximation, see `eval_dbr_unitary`).
        If $s>0$ then for $D = \Delta(H)$ the GWW DBR will give a $\sigma$-decrease,
        see https://arxiv.org/abs/2206.11772."""

        operator = self.eval_dbr_unitary(step, mode, d)
        operator_dagger = self.backend.cast(
            np.array(np.matrix(self.backend.to_numpy(operator)).getH())
        )
        self.h.matrix = operator_dagger @ self.h.matrix @ operator
        return operator

    def eval_dbr_unitary(
        self,
        step: float,
        mode: DoubleBracketGeneratorType = None,
        d: np.array = None,
    ):
        """In call we are working in the convention that $H' = U^\\dagger H
        U$ where $U=e^{-sW}$ with $W=[D,H]$ or an approximation of that by a group commutator.
        That is handy because if we switch from the DBI in the Heisenberg picture for the
        Hamiltonian, we get that the transformation of the state is $|\\psi'\rangle = U |\\psi\rangle$
        so that $\\langle H\rangle_{\\psi'} = \\langle H' \rangle_\\psi$ (i.e. when writing the unitary
        acting on the state dagger notation is avoided).
        The group commutator must approximate $U=e^{-s[D,H]}$. This is achieved by setting $r = \\sqrt{s}$ so that
        $$V = e^{-irH}e^{irD}e^{irH}e^{-irD}$$
        because
        $$e^{-irH}De^{irH} = D+ir[D,H]+O(r^2)$$
        so
        $$V\approx e^{irD +i^2 r^2[D,H] + O(r^2) -irD} \approx U\\ .$$
        See the app in https://arxiv.org/abs/2206.11772 for a derivation.
        """
        if mode is None:
            mode = self.mode

        if mode is DoubleBracketGeneratorType.canonical:
            operator = self.backend.calculate_matrix_exp(
                -1.0j * step,
                self.commutator(self.diagonal_h_matrix, self.h.matrix),
            )
        elif mode is DoubleBracketGeneratorType.single_commutator:
            if d is None:
                d = self.diagonal_h_matrix
            operator = self.backend.calculate_matrix_exp(
                -1.0j * step,
                self.commutator(self.backend.cast(d), self.h.matrix),
            )
        elif mode is DoubleBracketGeneratorType.group_commutator:
            if d is None:
                d = self.diagonal_h_matrix
            operator = (
                self.h.exp(step)
                @ self.backend.calculate_matrix_exp(-step, d)
                @ self.h.exp(-step)
                @ self.backend.calculate_matrix_exp(step, d)
            )
        elif mode is DoubleBracketGeneratorType.group_commutator_third_order:
            if d is None:
                d = self.diagonal_h_matrix
            operator = (
                self.h.exp(-step * (np.sqrt(5) - 1) / 2)
                @ self.backend.calculate_matrix_exp(-step * (np.sqrt(5) - 1) / 2, d)
                @ self.h.exp(step)
                @ self.backend.calculate_matrix_exp(step * (np.sqrt(5) + 1) / 2, d)
                @ self.h.exp(-step * (3 - np.sqrt(5)) / 2)
                @ self.backend.calculate_matrix_exp(-step, d)
            )
            operator = (
                self.backend.calculate_matrix_exp(step, d)
                @ self.h.exp(step * (3 - np.sqrt(5)) / 2)
                @ self.backend.calculate_matrix_exp(-step * (np.sqrt(5) + 1) / 2, d)
                @ self.h.exp(-step)
                @ self.backend.calculate_matrix_exp(step * (np.sqrt(5) - 1) / 2, d)
                @ self.h.exp(step * (np.sqrt(5) - 1) / 2)
            )
        else:
            raise NotImplementedError(
                f"The mode {mode} is not supported"
            )  # pragma: no cover

        return operator

    @staticmethod
    def commutator(a, b):
        """Compute commutator between two arrays."""
        return a @ b - b @ a

    @property
    def diagonal_h_matrix(self):
        """Diagonal H matrix."""
        return self.backend.cast(np.diag(np.diag(self.backend.to_numpy(self.h.matrix))))

    @property
    def off_diag_h(self):
        """Off-diagonal H matrix."""
        return self.h.matrix - self.diagonal_h_matrix

    @property
    def off_diagonal_norm(self):
        """Hilbert Schmidt norm of off-diagonal part of H matrix, namely :math:`\\text{Tr}(\\sqrt{A^{\\dagger} A})`."""
        off_diag_h_dag = self.backend.cast(
            np.matrix(self.backend.to_numpy(self.off_diag_h)).getH()
        )
        return np.sqrt(
            np.real(np.trace(self.backend.to_numpy(off_diag_h_dag @ self.off_diag_h)))
        )

    @property
    def backend(self):
        """Get Hamiltonian's backend."""
        return self.h0.backend

    @property
    def nqubits(self):
        """Number of qubits."""
        return self.h.nqubits

    def least_squares(self, d: np.array):
        """Least squares cost function."""
        d = self.backend.to_numpy(d)
        return np.real(
            0.5 * np.linalg.norm(d) ** 2
            - np.trace(self.backend.to_numpy(self.h.matrix) @ d)
        )

    def choose_step(
        self,
        d: Optional[np.array] = None,
        scheduling: Optional[DoubleBracketScheduling] = None,
        **kwargs,
    ):
        """Calculate the optimal step using respective the `scheduling` methods."""
        if scheduling is None:
            scheduling = self.scheduling
        step = scheduling(self, d=d, **kwargs)
        # TODO: write test for this case
        if (
            step is None
            and scheduling is DoubleBracketScheduling.polynomial_approximation
        ):  # pragma: no cover
            kwargs["n"] = kwargs.get("n", 3)
            kwargs["n"] += 1
            # if n==n_max, return None
            step = scheduling(self, d=d, **kwargs)
            # if for a given polynomial order n, no solution is found, we increase the order of the polynomial by 1
        return step

    def loss(self, step: float, d: np.array = None, look_ahead: int = 1):
        """
        Compute loss function distance between `look_ahead` steps.

        Args:
            step (float): iteration step.
            d (np.array): diagonal operator, use canonical by default.
            look_ahead (int): number of iteration steps to compute the loss function;
        """
        # copy initial hamiltonian
        h_copy = deepcopy(self.h)

        for _ in range(look_ahead):
            self.__call__(mode=self.mode, step=step, d=d)

        # loss values depending on the cost function
        if self.cost is DoubleBracketCostFunction.off_diagonal_norm:
            loss = self.off_diagonal_norm
        elif self.cost is DoubleBracketCostFunction.least_squares:
            loss = self.least_squares(d)
        elif self.cost == DoubleBracketCostFunction.energy_fluctuation:
            loss = self.energy_fluctuation(self.ref_state)

        # set back the initial configuration
        self.h = h_copy

        return loss

    def energy_fluctuation(self, state):
        """
        Evaluate energy fluctuation.

        .. math::
            \\Xi(\\mu) = \\sqrt{\\langle\\mu|\\hat{H}^2|\\mu\\rangle - \\langle\\mu|\\hat{H}|\\mu\\rangle^2} \\,

        for a given state :math:`|\\mu\\rangle`.

        Args:
            state (np.ndarray): quantum state to be used to compute the energy fluctuation with H.
        """
        return self.h.energy_fluctuation(state)

    def sigma(self, h: np.array):
        """Returns the off-diagonal restriction of matrix `h`."""
        return self.backend.cast(h) - self.backend.cast(
            np.diag(np.diag(self.backend.to_numpy(h)))
        )

    def generate_gamma_list(self, n: int, d: np.array):
        r"""Computes the n-nested Gamma functions, where $\Gamma_k=[W,...,[W,[W,H]]...]$, where we take k nested commutators with $W = [D, H]$"""
        W = self.commutator(self.backend.cast(d), self.sigma(self.h.matrix))
        gamma_list = [self.h.matrix]
        for _ in range(n - 1):
            gamma_list.append(self.commutator(W, gamma_list[-1]))
        return gamma_list

    def cost_expansion(self, d, n):
        d = self.backend.cast(d)

        if self.cost is DoubleBracketCostFunction.off_diagonal_norm:
            coef = off_diagonal_norm_polynomial_expansion_coef(self, d, n)
        elif self.cost is DoubleBracketCostFunction.least_squares:
            coef = least_squares_polynomial_expansion_coef(self, d, n)
        elif self.cost is DoubleBracketCostFunction.energy_fluctuation:
            coef = energy_fluctuation_polynomial_expansion_coef(
                self, d, n, self.ref_state
            )
        else:  # pragma: no cover
            raise ValueError(f"Cost function {self.cost} not recognized.")
        return coef



import math
from functools import partial
from typing import Optional

import hyperopt
import numpy as np

error = 1e-3


def grid_search_step(
    dbi_object,
    step_min: float = 1e-5,
    step_max: float = 1,
    num_evals: int = 100,
    space: Optional[np.array] = None,
    d: Optional[np.array] = None,
):
    """
    Greedy optimization of the iteration step.

    Args:
        step_min: lower bound of the search grid;
        step_max: upper bound of the search grid;
        mnum_evals: number of iterations between step_min and step_max;
        d: diagonal operator for generating double-bracket iterations.

    Returns:
        (float): optimized best iteration step (minimizing off-diagonal norm).
    """
    if space is None:
        space = np.linspace(step_min, step_max, num_evals)

    if d is None:
        d = dbi_object.diagonal_h_matrix

    loss_list = [dbi_object.loss(step, d=d) for step in space]

    idx_max_loss = np.argmin(loss_list)
    return space[idx_max_loss]


def hyperopt_step(
    dbi_object,
    step_min: float = 1e-5,
    step_max: float = 1,
    max_evals: int = 100,
    space: callable = None,
    optimizer: callable = None,
    look_ahead: int = 1,
    d: Optional[np.array] = None,
):
    """
    Optimize iteration step using hyperopt.

    Args:
        step_min: lower bound of the search grid;
        step_max: upper bound of the search grid;
        max_evals: maximum number of iterations done by the hyperoptimizer;
        space: see hyperopt.hp possibilities;
        optimizer: see hyperopt algorithms;
        look_ahead: number of iteration steps to compute the loss function;
        d: diagonal operator for generating double-bracket iterations.

    Returns:
        (float): optimized best iteration step (minimizing loss function).
    """
    if space is None:
        space = hyperopt.hp.uniform
    if optimizer is None:
        optimizer = hyperopt.tpe
    if d is None:
        d = dbi_object.diagonal_h_matrix

    space = space("step", step_min, step_max)

    best = hyperopt.fmin(
        fn=partial(dbi_object.loss, d=d, look_ahead=look_ahead),
        space=space,
        algo=optimizer.suggest,
        max_evals=max_evals,
        show_progressbar=False,
    )
    return best["step"]


def polynomial_step(
    dbi_object,
    n: int = 2,
    n_max: int = 5,
    d: np.array = None,
    coef: Optional[list] = None,
    cost: Optional[str] = None,
):
    r"""
    Optimizes iteration step by solving the n_th order polynomial expansion of the loss function.
    e.g. $n=2$: $2\Trace(\sigma(\Gamma_1 + s\Gamma_2 + s^2/2\Gamma_3)\sigma(\Gamma_0 + s\Gamma_1 + s^2/2\Gamma_2))
    Args:
        n (int, optional): the order to which the loss function is expanded. Defaults to 4.
        n_max (int, optional): maximum order allowed for recurring calls of `polynomial_step`. Defaults to 5.
        d (np.array, optional): diagonal operator, default as $\delta(H)$.
        backup_scheduling (`DoubleBracketScheduling`): the scheduling method to use in case no real positive roots are found.
    """
    if cost is None:
        cost = dbi_object.cost

    if d is None:
        d = dbi_object.diagonal_h_matrix

    if n > n_max:
        raise ValueError(
            "No solution can be found with polynomial approximation. Increase `n_max` or use other scheduling methods."
        )
    if coef is None:
        coef = dbi_object.cost_expansion(d=d, n=n)
    roots = np.roots(coef)
    real_positive_roots = [
        np.real(root) for root in roots if np.imag(root) < 1e-3 and np.real(root) > 0
    ]
    # solution exists, return minimum s
    if len(real_positive_roots) > 0:
        losses = [dbi_object.loss(step=root, d=d) for root in real_positive_roots]
        return real_positive_roots[losses.index(min(losses))]
    # solution does not exist, return None
    else:
        return None


def simulated_annealing_step(
    dbi_object,
    d: Optional[np.array] = None,
    initial_s=None,
    step_min=1e-5,
    step_max=1,
    s_jump_range=None,
    s_jump_range_divident=5,
    initial_temp=1,
    cooling_rate=0.85,
    min_temp=1e-5,
    max_iter=200,
):
    """
    Perform a single step of simulated annealing optimization.

    Parameters:
        dbi_object: DBI object
            The object representing the problem to be optimized.
        d: Optional[np.array], optional
            The diagonal matrix 'd' used in optimization. If None, it uses the diagonal
            matrix 'diagonal_h_matrix' from dbi_object.
        initial_s: float or None, optional
            Initial value for 's', the step size. If None, it is initialized using
            polynomial_step function with 'n=4'. If 'polynomial_step' returns None,
            'initial_s' is set to 'step_min'.
        step_min: float, optional
            Minimum value for the step size 's'.
        step_max: float, optional
            Maximum value for the step size 's'.
        s_jump_range: float or None, optional
            Range for the random jump in step size. If None, it's calculated based on
            'step_min', 'step_max', and 's_jump_range_divident'.
        s_jump_range_divident: int, optional
            Dividend to determine the range for random jump in step size.
        initial_temp: float, optional
            Initial temperature for simulated annealing.
        cooling_rate: float, optional
            Rate at which temperature decreases in simulated annealing.
        min_temp: float, optional
            Minimum temperature threshold for termination of simulated annealing.
        max_iter: int, optional
            Maximum number of iterations for simulated annealing.

    Returns:
        float:
            The optimized step size 's'.
    """

    if d is None:
        d = dbi_object.diagonal_h_matrix
    if initial_s is None:
        initial_s = polynomial_step(dbi_object=dbi_object, d=d, n=4)
        # TODO: implement test to catch this if statement
        if initial_s is None:  # pragma: no cover
            initial_s = step_min
    if s_jump_range is None:
        s_jump_range = (step_max - step_min) / s_jump_range_divident
    current_s = initial_s
    current_loss = dbi_object.loss(d=d, step=current_s)
    temp = initial_temp

    for _ in range(max_iter):
        candidate_s = max(
            step_min,
            min(
                current_s + np.random.uniform(-1 * s_jump_range, s_jump_range), step_max
            ),
        )
        candidate_loss = dbi_object.loss(d=d, step=candidate_s)

        # Calculate change in loss
        delta_loss = candidate_loss - current_loss

        # Determine if the candidate solution is an improvement
        if delta_loss < 0 or np.random.rand() < math.exp(-delta_loss / temp):
            current_s = candidate_s
            current_loss = candidate_loss
        # Cool down
        temp *= cooling_rate
        if temp < min_temp or current_s > step_max or current_s < step_min:
            break

    return current_s



from functools import cache, reduce

import numpy as np
from scipy import sparse

name = "numpy"


def _get_rxz(symplectic_matrix, nqubits):
    return (
        symplectic_matrix[:, -1],
        symplectic_matrix[:, :nqubits],
        symplectic_matrix[:, nqubits:-1],
    )


def I(symplectic_matrix, q, nqubits):
    return symplectic_matrix


def H(symplectic_matrix, q, nqubits):
    r, x, z = _get_rxz(symplectic_matrix, nqubits)
    symplectic_matrix[:, -1] = r ^ (x[:, q] & z[:, q])
    symplectic_matrix[:, [q, nqubits + q]] = symplectic_matrix[:, [nqubits + q, q]]
    return symplectic_matrix


def CNOT(symplectic_matrix, control_q, target_q, nqubits):
    ind_zt = nqubits + target_q
    ind_zc = nqubits + control_q
    r = symplectic_matrix[:, -1]
    xcq = symplectic_matrix[:, control_q]
    xtq = symplectic_matrix[:, target_q]
    ztq = symplectic_matrix[:, ind_zt]
    zcq = symplectic_matrix[:, ind_zc]
    symplectic_matrix[:, -1] = r ^ (xcq & ztq) & (xtq ^ ~zcq)
    symplectic_matrix[:, target_q] = xtq ^ xcq
    symplectic_matrix[:, ind_zc] = zcq ^ ztq
    return symplectic_matrix


def CZ(symplectic_matrix, control_q, target_q, nqubits):
    """Decomposition --> H-CNOT-H"""
    ind_zt = nqubits + target_q
    ind_zc = nqubits + control_q
    r = symplectic_matrix[:, -1]
    xcq = symplectic_matrix[:, control_q]
    xtq = symplectic_matrix[:, target_q]
    ztq = symplectic_matrix[:, ind_zt]
    zcq = symplectic_matrix[:, ind_zc]
    ztq_xor_xcq = ztq ^ xcq
    symplectic_matrix[:, -1] = (
        r ^ (xtq & ztq) ^ (xcq & xtq & (ztq ^ ~zcq)) ^ (xtq & ztq_xor_xcq)
    )
    z_control_q = xtq ^ zcq
    z_target_q = ztq_xor_xcq
    symplectic_matrix[:, ind_zc] = z_control_q
    symplectic_matrix[:, ind_zt] = z_target_q
    return symplectic_matrix


def S(symplectic_matrix, q, nqubits):
    r, x, z = _get_rxz(symplectic_matrix, nqubits)
    r[:] = r ^ (x[:, q] & z[:, q])
    z[:, q] = z[:, q] ^ x[:, q]
    return symplectic_matrix


def Z(symplectic_matrix, q, nqubits):
    """Decomposition --> S-S"""
    r, x, z = _get_rxz(symplectic_matrix, nqubits)
    symplectic_matrix[:, -1] = r ^ ((x[:, q] & z[:, q]) ^ x[:, q] & (z[:, q] ^ x[:, q]))
    return symplectic_matrix


def X(symplectic_matrix, q, nqubits):
    """Decomposition --> H-S-S-H"""
    r, x, z = _get_rxz(symplectic_matrix, nqubits)
    symplectic_matrix[:, -1] = r ^ (z[:, q] & (z[:, q] ^ x[:, q])) ^ (z[:, q] & x[:, q])
    return symplectic_matrix


def Y(symplectic_matrix, q, nqubits):
    """Decomposition --> S-S-H-S-S-H"""
    r, x, z = _get_rxz(symplectic_matrix, nqubits)
    symplectic_matrix[:, -1] = (
        r ^ (z[:, q] & (z[:, q] ^ x[:, q])) ^ (x[:, q] & (z[:, q] ^ x[:, q]))
    )
    return symplectic_matrix


def SX(symplectic_matrix, q, nqubits):
    """Decomposition --> H-S-H"""
    r, x, z = _get_rxz(symplectic_matrix, nqubits)
    symplectic_matrix[:, -1] = r ^ (z[:, q] & (z[:, q] ^ x[:, q]))
    symplectic_matrix[:, q] = z[:, q] ^ x[:, q]
    return symplectic_matrix


def SDG(symplectic_matrix, q, nqubits):
    """Decomposition --> S-S-S"""
    r, x, z = _get_rxz(symplectic_matrix, nqubits)
    symplectic_matrix[:, -1] = r ^ (x[:, q] & (z[:, q] ^ x[:, q]))
    symplectic_matrix[:, nqubits + q] = z[:, q] ^ x[:, q]
    return symplectic_matrix


def SXDG(symplectic_matrix, q, nqubits):
    """Decomposition --> H-S-S-S-H"""
    r, x, z = _get_rxz(symplectic_matrix, nqubits)
    symplectic_matrix[:, -1] = r ^ (z[:, q] & x[:, q])
    symplectic_matrix[:, q] = z[:, q] ^ x[:, q]
    return symplectic_matrix


def RX(symplectic_matrix, q, nqubits, theta):
    if theta % (2 * np.pi) == 0:
        return I(symplectic_matrix, q, nqubits)
    elif (theta / np.pi - 1) % 2 == 0:
        return X(symplectic_matrix, q, nqubits)
    elif (theta / (np.pi / 2) - 1) % 4 == 0:
        return SX(symplectic_matrix, q, nqubits)
    else:  # theta == 3*pi/2 + 2*n*pi
        return SXDG(symplectic_matrix, q, nqubits)


def RZ(symplectic_matrix, q, nqubits, theta):
    if theta % (2 * np.pi) == 0:
        return I(symplectic_matrix, q, nqubits)
    elif (theta / np.pi - 1) % 2 == 0:
        return Z(symplectic_matrix, q, nqubits)
    elif (theta / (np.pi / 2) - 1) % 4 == 0:
        return S(symplectic_matrix, q, nqubits)
    else:  # theta == 3*pi/2 + 2*n*pi
        return SDG(symplectic_matrix, q, nqubits)


def RY_pi(symplectic_matrix, q, nqubits):
    """Decomposition --> H-S-S"""
    r, x, z = _get_rxz(symplectic_matrix, nqubits)
    symplectic_matrix[:, -1] = r ^ (x[:, q] & (z[:, q] ^ x[:, q]))
    symplectic_matrix[:, [nqubits + q, q]] = symplectic_matrix[:, [q, nqubits + q]]
    return symplectic_matrix


def RY_3pi_2(symplectic_matrix, q, nqubits):
    """Decomposition --> H-S-S-H-S-S-H-S-S"""
    r, x, z = _get_rxz(symplectic_matrix, nqubits)
    symplectic_matrix[:, -1] = r ^ (z[:, q] & (z[:, q] ^ x[:, q]))
    symplectic_matrix[:, [nqubits + q, q]] = symplectic_matrix[:, [q, nqubits + q]]
    return symplectic_matrix


def RY(symplectic_matrix, q, nqubits, theta):
    if theta % (2 * np.pi) == 0:
        return I(symplectic_matrix, q, nqubits)
    elif (theta / np.pi - 1) % 2 == 0:
        return Y(symplectic_matrix, q, nqubits)
    elif (theta / (np.pi / 2) - 1) % 4 == 0:
        """Decomposition --> H-S-S"""
        return RY_pi(symplectic_matrix, q, nqubits)
    else:  # theta == 3*pi/2 + 2*n*pi
        """Decomposition --> H-S-S-H-S-S-H-S-S"""
        return RY_3pi_2(symplectic_matrix, q, nqubits)


def SWAP(symplectic_matrix, control_q, target_q, nqubits):
    """Decomposition --> CNOT-CNOT-CNOT"""
    r, x, z = _get_rxz(symplectic_matrix, nqubits)
    symplectic_matrix[:, -1] = (
        r
        ^ (x[:, control_q] & z[:, target_q] & (x[:, target_q] ^ ~z[:, control_q]))
        ^ (
            (x[:, target_q] ^ x[:, control_q])
            & (z[:, target_q] ^ z[:, control_q])
            & (z[:, target_q] ^ ~x[:, control_q])
        )
        ^ (
            x[:, target_q]
            & z[:, control_q]
            & (x[:, control_q] ^ x[:, target_q] ^ z[:, control_q] ^ ~z[:, target_q])
        )
    )
    symplectic_matrix[
        :, [control_q, target_q, nqubits + control_q, nqubits + target_q]
    ] = symplectic_matrix[
        :, [target_q, control_q, nqubits + target_q, nqubits + control_q]
    ]
    return symplectic_matrix


def iSWAP(symplectic_matrix, control_q, target_q, nqubits):
    """Decomposition --> H-CNOT-CNOT-H-S-S"""
    r, x, z = _get_rxz(symplectic_matrix, nqubits)
    symplectic_matrix[:, -1] = (
        r
        ^ (x[:, target_q] & z[:, target_q])
        ^ (x[:, control_q] & z[:, control_q])
        ^ (x[:, control_q] & (z[:, control_q] ^ x[:, control_q]))
        ^ (
            (z[:, control_q] ^ x[:, control_q])
            & (z[:, target_q] ^ x[:, target_q])
            & (x[:, target_q] ^ ~x[:, control_q])
        )
        ^ (
            (x[:, target_q] ^ z[:, control_q] ^ x[:, control_q])
            & (x[:, target_q] ^ z[:, target_q] ^ x[:, control_q])
            & (x[:, target_q] ^ z[:, target_q] ^ x[:, control_q] ^ ~z[:, control_q])
        )
        ^ (x[:, control_q] & (x[:, target_q] ^ x[:, control_q] ^ z[:, control_q]))
    )
    z_control_q = x[:, target_q] ^ z[:, target_q] ^ x[:, control_q]
    z_target_q = x[:, target_q] ^ z[:, control_q] ^ x[:, control_q]
    symplectic_matrix[:, nqubits + control_q] = z_control_q
    symplectic_matrix[:, nqubits + target_q] = z_target_q
    symplectic_matrix[:, [control_q, target_q]] = symplectic_matrix[
        :, [target_q, control_q]
    ]
    return symplectic_matrix


def FSWAP(symplectic_matrix, control_q, target_q, nqubits):
    """Decomposition --> X-CNOT-RY-CNOT-RY-CNOT-CNOT-X"""
    symplectic_matrix = X(symplectic_matrix, target_q, nqubits)
    symplectic_matrix = CNOT(symplectic_matrix, control_q, target_q, nqubits)
    symplectic_matrix = RY(symplectic_matrix, control_q, nqubits, np.pi / 2)
    symplectic_matrix = CNOT(symplectic_matrix, target_q, control_q, nqubits)
    symplectic_matrix = RY(symplectic_matrix, control_q, nqubits, -np.pi / 2)
    symplectic_matrix = CNOT(symplectic_matrix, target_q, control_q, nqubits)
    symplectic_matrix = CNOT(symplectic_matrix, control_q, target_q, nqubits)
    return X(symplectic_matrix, control_q, nqubits)


def CY(symplectic_matrix, control_q, target_q, nqubits):
    """Decomposition --> S-CNOT-SDG"""
    r, x, z = _get_rxz(symplectic_matrix, nqubits)
    symplectic_matrix[:, -1] = (
        r
        ^ (x[:, target_q] & (z[:, target_q] ^ x[:, target_q]))
        ^ (
            x[:, control_q]
            & (x[:, target_q] ^ z[:, target_q])
            & (z[:, control_q] ^ ~x[:, target_q])
        )
        ^ ((x[:, target_q] ^ x[:, control_q]) & (z[:, target_q] ^ x[:, target_q]))
    )
    x_target_q = x[:, control_q] ^ x[:, target_q]
    z_control_q = z[:, control_q] ^ z[:, target_q] ^ x[:, target_q]
    z_target_q = z[:, target_q] ^ x[:, control_q]
    symplectic_matrix[:, target_q] = x_target_q
    symplectic_matrix[:, nqubits + control_q] = z_control_q
    symplectic_matrix[:, nqubits + target_q] = z_target_q
    return symplectic_matrix


def CRX(symplectic_matrix, control_q, target_q, nqubits, theta):
    # theta = 4 * n * pi
    if theta % (4 * np.pi) == 0:
        return I(symplectic_matrix, target_q, nqubits)
    # theta = pi + 4 * n * pi
    elif (theta / np.pi - 1) % 4 == 0:
        symplectic_matrix = X(symplectic_matrix, target_q, nqubits)
        symplectic_matrix = CZ(symplectic_matrix, control_q, target_q, nqubits)
        symplectic_matrix = X(symplectic_matrix, target_q, nqubits)
        return CY(symplectic_matrix, control_q, target_q, nqubits)
    # theta = 2 * pi + 4 * n * pi
    elif (theta / (2 * np.pi) - 1) % 2 == 0:
        symplectic_matrix = CZ(symplectic_matrix, control_q, target_q, nqubits)
        symplectic_matrix = Y(symplectic_matrix, target_q, nqubits)
        symplectic_matrix = CZ(symplectic_matrix, control_q, target_q, nqubits)
        return Y(symplectic_matrix, target_q, nqubits)
    # theta = 3 * pi + 4 * n * pi
    elif (theta / np.pi - 3) % 4 == 0:
        symplectic_matrix = X(symplectic_matrix, target_q, nqubits)
        symplectic_matrix = CY(symplectic_matrix, control_q, target_q, nqubits)
        symplectic_matrix = X(symplectic_matrix, target_q, nqubits)
        return CZ(symplectic_matrix, control_q, target_q, nqubits)


def CRZ(symplectic_matrix, control_q, target_q, nqubits, theta):
    # theta = 4 * n * pi
    if theta % (4 * np.pi) == 0:
        return I(symplectic_matrix, target_q, nqubits)
    # theta = pi + 4 * n * pi
    elif (theta / np.pi - 1) % 4 == 0:
        symplectic_matrix = X(symplectic_matrix, target_q, nqubits)
        symplectic_matrix = CY(symplectic_matrix, control_q, target_q, nqubits)
        symplectic_matrix = X(symplectic_matrix, target_q, nqubits)
        return CNOT(symplectic_matrix, control_q, target_q, nqubits)
    # theta = 2 * pi + 4 * n * pi
    elif (theta / (2 * np.pi) - 1) % 2 == 0:
        symplectic_matrix = CZ(symplectic_matrix, control_q, target_q, nqubits)
        symplectic_matrix = X(symplectic_matrix, target_q, nqubits)
        symplectic_matrix = CZ(symplectic_matrix, control_q, target_q, nqubits)
        return X(symplectic_matrix, target_q, nqubits)
    # theta = 3 * pi + 4 * n * pi
    elif (theta / np.pi - 3) % 4 == 0:
        symplectic_matrix = CNOT(symplectic_matrix, control_q, target_q, nqubits)
        symplectic_matrix = X(symplectic_matrix, target_q, nqubits)
        symplectic_matrix = CY(symplectic_matrix, control_q, target_q, nqubits)
        return X(symplectic_matrix, target_q, nqubits)


def CRY(symplectic_matrix, control_q, target_q, nqubits, theta):
    # theta = 4 * n * pi
    if theta % (4 * np.pi) == 0:
        return I(symplectic_matrix, target_q, nqubits)
    # theta = pi + 4 * n * pi
    elif (theta / np.pi - 1) % 4 == 0:
        symplectic_matrix = Z(symplectic_matrix, target_q, nqubits)
        symplectic_matrix = CNOT(symplectic_matrix, control_q, target_q, nqubits)
        symplectic_matrix = Z(symplectic_matrix, target_q, nqubits)
        return CZ(symplectic_matrix, control_q, target_q, nqubits)
    # theta = 2 * pi + 4 * n * pi
    elif (theta / (2 * np.pi) - 1) % 2 == 0:
        return CRZ(symplectic_matrix, control_q, target_q, nqubits, theta)
    # theta = 3 * pi + 4 * n * pi
    elif (theta / np.pi - 3) % 4 == 0:
        symplectic_matrix = CZ(symplectic_matrix, control_q, target_q, nqubits)
        symplectic_matrix = Z(symplectic_matrix, target_q, nqubits)
        symplectic_matrix = CNOT(symplectic_matrix, control_q, target_q, nqubits)
        return Z(symplectic_matrix, target_q, nqubits)


def ECR(symplectic_matrix, control_q, target_q, nqubits):
    symplectic_matrix = S(symplectic_matrix, control_q, nqubits)
    symplectic_matrix = SX(symplectic_matrix, target_q, nqubits)
    symplectic_matrix = CNOT(symplectic_matrix, control_q, target_q, nqubits)
    return X(symplectic_matrix, control_q, nqubits)


def _exponent(
    x1: np.ndarray, z1: np.ndarray, x2: np.ndarray, z2: np.ndarray
) -> np.ndarray:
    """Helper function that computes the exponent to which i is raised for the product of the x and z paulis encoded in the symplectic matrix. This is used in _rowsum. The computation is performed parallely over the separated paulis x1[i], z1[i], x2[i] and z2[i].

    Args:
        x1 (np.array): Bits of the first x paulis.
        z1 (np.array): Bits of the first z paulis.
        x2 (np.array): Bits of the second x paulis.
        z2 (np.array): Bits of the second z paulis.

    Returns:
        (np.array): The calculated exponents.
    """
    return 2 * (x1 * x2 * (z2 - z1) + z1 * z2 * (x1 - x2)) - x1 * z2 + x2 * z1


def _rowsum(symplectic_matrix, h, i, nqubits, determined=False):
    """Helper function that updates the symplectic matrix by setting the h-th generator equal to the (i+h)-th one. This is done to keep track of the phase of the h-th row of the symplectic matrix (r[h]). The function is applied parallely over all the rows h and i passed.

    Args:
        symplectic_matrix (np.array): Input symplectic matrix.
        h (np.array): Indices of the rows encoding the generators to update.
        i (np.array): Indices of the rows encoding the generators to use.
        nqubits (int): Total number of qubits.

    Returns:
        (np.array): The updated symplectic matrix.
    """
    xi, zi = symplectic_matrix[i, :nqubits], symplectic_matrix[i, nqubits:-1]
    xh, zh = symplectic_matrix[h, :nqubits], symplectic_matrix[h, nqubits:-1]
    exponents = _exponent(xi, zi, xh, zh)
    ind = (
        2 * symplectic_matrix[h, -1]
        + 2 * symplectic_matrix[i, -1]
        + np.sum(exponents, axis=-1)
    ) % 4 == 0
    r = np.ones(h.shape[0], dtype=np.uint8)
    r[ind] = 0

    xi_xh = xi ^ xh
    zi_zh = zi ^ zh
    if determined:
        r = reduce(np.logical_xor, r)
        xi_xh = reduce(np.logical_xor, xi_xh)
        zi_zh = reduce(np.logical_xor, zi_zh)
        symplectic_matrix[h[0], -1] = r
        symplectic_matrix[h[0], :nqubits] = xi_xh
        symplectic_matrix[h[0], nqubits:-1] = zi_zh
    else:
        symplectic_matrix[h, -1] = r
        symplectic_matrix[h, :nqubits] = xi_xh
        symplectic_matrix[h, nqubits:-1] = zi_zh
    return symplectic_matrix


def _determined_outcome(state, q, nqubits):
    state[-1, :] = 0
    idx = (state[:nqubits, q].nonzero()[0] + nqubits).astype(np.uint)
    state = _pack_for_measurements(state, nqubits)
    state = _rowsum(
        state,
        2 * nqubits * np.ones(idx.shape, dtype=np.uint),
        idx,
        _packed_size(nqubits),
        True,
    )
    state = _unpack_for_measurements(state, nqubits)
    return state, state[-1, -1]


def _random_outcome(state, p, q, nqubits):
    p = p[0] + nqubits
    tmp = state[p, q].copy()
    state[p, q] = 0
    h = state[:-1, q].nonzero()[0]
    state[p, q] = tmp
    if h.shape[0] > 0:
        state = _pack_for_measurements(state, nqubits)
        state = _rowsum(
            state,
            h.astype(np.uint),
            p * np.ones(h.shape[0], dtype=np.uint),
            _packed_size(nqubits),
            False,
        )
        state = _unpack_for_measurements(state, nqubits)
    state[p - nqubits, :] = state[p, :]
    outcome = np.random.randint(2, size=1).item()
    state[p, :] = 0
    state[p, -1] = outcome
    state[p, nqubits + q] = 1
    return state, outcome


@cache
def _dim(nqubits):
    """Returns the dimension of the symplectic matrix for a given number of qubits."""
    return 2 * nqubits + 1


@cache
def _packed_size(n):
    """Returns the size of an array of `n` booleans after packing."""
    return np.ceil(n / 8).astype(int)


def _packbits(array, axis):
    return np.packbits(array, axis=axis)


def _unpackbits(array, axis):
    return np.unpackbits(array, axis=axis)


def _pack_for_measurements(state, nqubits):
    """Prepares the state for measurements by packing the rows of the X and Z sections of the symplectic matrix."""
    r, x, z = _get_rxz(state, nqubits)
    x = _packbits(x, axis=1)
    z = _packbits(z, axis=1)
    return np.hstack((x, z, r[:, None]))


@cache
def _pad_size(n):
    """Returns the size of the pad added to an array of original dimension `n` after unpacking."""
    return 8 - (n % 8)


def _unpack_for_measurements(state, nqubits):
    """Unpacks the symplectc matrix that was packed for measurements."""
    xz = _unpackbits(state[:, :-1], axis=1)
    padding_size = _pad_size(nqubits)
    x, z = xz[:, :nqubits], xz[:, nqubits + padding_size : -padding_size]
    return np.hstack((x, z, state[:, -1][:, None]))


def _init_state_for_measurements(state, nqubits, collapse):
    if collapse:
        return _unpackbits(state, axis=0)[: _dim(nqubits)]
    else:
        return state.copy()


# valid for a standard basis measurement only
def M(state, qubits, nqubits, collapse=False):
    sample = []
    state = _init_state_for_measurements(state, nqubits, collapse)
    for q in qubits:
        p = state[nqubits:-1, q].nonzero()[0]
        # random outcome, affects the state
        if len(p) > 0:
            state, outcome = _random_outcome(state, p, q, nqubits)
        # determined outcome, state unchanged
        else:
            state, outcome = _determined_outcome(state, q, nqubits)
        sample.append(outcome)
    if collapse:
        state = _packbits(state, axis=0)
    return sample


def cast(x, dtype=None, copy=False):
    if dtype is None:
        dtype = "complex128"
    if isinstance(x, np.ndarray):
        return x.astype(dtype, copy=copy)
    elif sparse.issparse(x):  # pragma: no cover
        return x.astype(dtype, copy=copy)
    return np.array(x, dtype=dtype, copy=copy)


def _clifford_pre_execution_reshape(state):
    """Reshape and packing applied to the symplectic matrix before execution to prepare the state in the form needed by each engine.

    Args:
        state (np.array): Input state.

    Returns:
        (np.array) The packed and reshaped state.
    """
    return _packbits(state, axis=0)


def _clifford_post_execution_reshape(state, nqubits: int):
    """Reshape and unpacking applied to the state after execution to retrieve the standard symplectic matrix form.

    Args:
        state (np.array): Input state.
        nqubits (int): Number of qubits.

    Returns:
        (np.array) The unpacked and reshaped state.
    """
    state = _unpackbits(state, axis=0)[: _dim(nqubits)]
    return state


def identity_density_matrix(nqubits, normalize: bool = True):
    state = np.eye(2**nqubits, dtype="complex128")
    if normalize is True:  # pragma: no cover
        state /= 2**nqubits
    return state



"""Module with functions that encode classical data into quantum circuits."""

import math
from inspect import signature
from typing import Optional, Union

import numpy as np

from qibo import gates
from qibo.config import raise_error
from qibo.models.circuit import Circuit


def comp_basis_encoder(
    basis_element: Union[int, str, list, tuple], nqubits: Optional[int] = None
):
    """Creates circuit that performs encoding of bitstrings into computational basis states.

    Args:
        basis_element (int or str or list or tuple): bitstring to be encoded.
            If ``int``, ``nqubits`` must be specified.
            If ``str``, must be composed of only :math:`0`s and :math:`1`s.
            If ``list`` or ``tuple``, must be composed of :math:`0`s and
            :math:`1`s as ``int`` or ``str``.
        nqubits (int, optional): total number of qubits in the circuit.
            If ``basis_element`` is ``int``, ``nqubits`` must be specified.
            If ``nqubits`` is ``None``, ``nqubits`` defaults to length of ``basis_element``.
            Defaults to ``None``.

    Returns:
       :class:`qibo.models.circuit.Circuit`: circuit encoding computational basis element.
    """
    if not isinstance(basis_element, (int, str, list, tuple)):
        raise_error(
            TypeError,
            "basis_element must be either type int or str or list or tuple, "
            + f"but it is type {type(basis_element)}.",
        )

    if isinstance(basis_element, (str, list, tuple)):
        if any(elem not in ["0", "1", 0, 1] for elem in basis_element):
            raise_error(ValueError, "all elements must be 0 or 1.")

    if nqubits is not None and not isinstance(nqubits, int):
        raise_error(
            TypeError, f"nqubits must be type int, but it is type {type(nqubits)}."
        )

    if nqubits is None:
        if isinstance(basis_element, int):
            raise_error(
                ValueError, f"nqubits must be specified when basis_element is type int."
            )
        else:
            nqubits = len(basis_element)

    if isinstance(basis_element, int):
        basis_element = f"{basis_element:0{nqubits}b}"

    if isinstance(basis_element, (str, tuple)):
        basis_element = list(basis_element)

    basis_element = list(map(int, basis_element))

    circuit = Circuit(nqubits)
    for qubit, elem in enumerate(basis_element):
        if elem == 1:
            circuit.add(gates.X(qubit))

    return circuit


def phase_encoder(data, rotation: str = "RY"):
    """Creates circuit that performs the phase encoding of ``data``.

    Args:
        data (ndarray or list): :math:`1`-dimensional array of phases to be loaded.
        rotation (str, optional): If ``"RX"``, uses :class:`qibo.gates.gates.RX` as rotation.
            If ``"RY"``, uses :class:`qibo.gates.gates.RY` as rotation.
            If ``"RZ"``, uses :class:`qibo.gates.gates.RZ` as rotation.
            Defaults to ``"RY"``.

    Returns:
        :class:`qibo.models.circuit.Circuit`: circuit that loads ``data`` in phase encoding.
    """
    if isinstance(data, list):
        data = np.array(data)

    if len(data.shape) != 1:
        raise_error(
            TypeError,
            f"``data`` must be a 1-dimensional array, but it has dimensions {data.shape}.",
        )

    if not isinstance(rotation, str):
        raise_error(
            TypeError,
            f"``rotation`` must be type str, but it is type {type(rotation)}.",
        )

    if rotation not in ["RX", "RY", "RZ"]:
        raise_error(ValueError, f"``rotation`` {rotation} not found.")

    nqubits = len(data)
    gate = getattr(gates, rotation.upper())

    circuit = Circuit(nqubits)
    circuit.add(gate(qubit, 0.0) for qubit in range(nqubits))
    circuit.set_parameters(data)

    return circuit


def unary_encoder(data, architecture: str = "tree"):
    """Creates circuit that performs the (deterministic) unary encoding of ``data``.

    Args:
        data (ndarray): :math:`1`-dimensional array of data to be loaded.
        architecture(str, optional): circuit architecture used for the unary loader.
            If ``diagonal``, uses a ladder-like structure.
            If ``tree``, uses a binary-tree-based structure.
            Defaults to ``tree``.

    Returns:
        :class:`qibo.models.circuit.Circuit`: circuit that loads ``data`` in unary representation.
    """
    if isinstance(data, list):
        data = np.array(data)

    if len(data.shape) != 1:
        raise_error(
            TypeError,
            f"``data`` must be a 1-dimensional array, but it has dimensions {data.shape}.",
        )

    if not isinstance(architecture, str):
        raise_error(
            TypeError,
            f"``architecture`` must be type str, but it is type {type(architecture)}.",
        )

    if architecture not in ["diagonal", "tree"]:
        raise_error(ValueError, f"``architecture`` {architecture} not found.")

    if architecture == "tree" and not math.log2(data.shape[0]).is_integer():
        raise_error(
            ValueError,
            "When ``architecture = 'tree'``, len(data) must be a power of 2. "
            + f"However, it is {len(data)}.",
        )

    nqubits = len(data)

    circuit = Circuit(nqubits)
    circuit.add(gates.X(nqubits - 1))
    circuit_rbs, _ = _generate_rbs_pairs(nqubits, architecture=architecture)
    circuit += circuit_rbs

    # calculating phases and setting circuit parameters
    phases = _generate_rbs_angles(data, nqubits, architecture)
    circuit.set_parameters(phases)

    return circuit


def unary_encoder_random_gaussian(nqubits: int, architecture: str = "tree", seed=None):
    """Creates a circuit that performs the unary encoding of a random Gaussian state.

    At depth :math:`h` of the tree architecture, the angles :math:`\\theta_{k} \\in [0, 2\\pi]` of the the
    gates :math:`RBS(\\theta_{k})` are sampled from the following probability density function:

    .. math::
        p_{h}(\\theta) = \\frac{1}{2} \\, \\frac{\\Gamma(2^{h-1})}{\\Gamma^{2}(2^{h-2})} \\,
            \\left|\\sin(\\theta) \\, \\cos(\\theta)\\right|^{2^{h-1} - 1} \\, ,

    where :math:`\\Gamma(\\cdot)` is the
    `Gamma function <https://en.wikipedia.org/wiki/Gamma_function>`_.

    Args:
        nqubits (int): number of qubits.
        architecture(str, optional): circuit architecture used for the unary loader.
            If ``tree``, uses a binary-tree-based structure.
            Defaults to ``tree``.
        seed (int or :class:`numpy.random.Generator`, optional): Either a generator of
            random numbers or a fixed seed to initialize a generator. If ``None``,
            initializes a generator with a random seed. Defaults to ``None``.

    Returns:
        :class:`qibo.models.circuit.Circuit`: circuit that loads a random Gaussian array in unary representation.

    References:
        1. A. Bouland, A. Dandapani, and A. Prakash, *A quantum spectral method for simulating
        stochastic processes, with applications to Monte Carlo*.
        `arXiv:2303.06719v1 [quant-ph] <https://arxiv.org/abs/2303.06719>`_
    """
    if not isinstance(nqubits, int):
        raise_error(
            TypeError, f"nqubits must be type int, but it is type {type(nqubits)}."
        )

    if nqubits <= 0.0:
        raise_error(
            ValueError, f"nqubits must be a positive integer, but it is {nqubits}."
        )

    if not isinstance(architecture, str):
        raise_error(
            TypeError,
            f"``architecture`` must be type str, but it is type {type(architecture)}.",
        )

    if architecture != "tree":
        raise_error(
            NotImplementedError,
            f"Currently, this function only accepts ``architecture=='tree'``.",
        )

    if not math.log2(nqubits).is_integer():
        raise_error(ValueError, f"nqubits must be a power of 2, but it is {nqubits}.")

    if (
        seed is not None
        and not isinstance(seed, int)
        and not isinstance(seed, np.random.Generator)
    ):
        raise_error(
            TypeError, "seed must be either type int or numpy.random.Generator."
        )

    from qibo.quantum_info.random_ensembles import (  # pylint: disable=C0415
        _ProbabilityDistributionGaussianLoader,
    )

    local_state = (
        np.random.default_rng(seed) if seed is None or isinstance(seed, int) else seed
    )

    sampler = _ProbabilityDistributionGaussianLoader(
        a=0, b=2 * math.pi, seed=local_state
    )

    circuit = Circuit(nqubits)
    circuit.add(gates.X(nqubits - 1))
    circuit_rbs, pairs_rbs = _generate_rbs_pairs(nqubits, architecture)
    circuit += circuit_rbs

    phases = []
    for depth, row in enumerate(pairs_rbs, 1):
        phases.extend(sampler.rvs(depth=depth, size=len(row)))

    circuit.set_parameters(phases)

    return circuit


def entangling_layer(
    nqubits: int,
    architecture: str = "diagonal",
    entangling_gate: Union[str, gates.Gate] = "CNOT",
    closed_boundary: bool = False,
):
    """Creates a layer of two-qubit, entangling gates.

    If the chosen gate is a parametrized gate, all phases are set to :math:`0.0`.

    Args:
        nqubits (int): Total number of qubits in the circuit.
        architecture (str, optional): Architecture of the entangling layer.
            Options are ``diagonal``, ``shifted``, ``even-layer``, and ``odd-layer``.
            Defaults to ``"diagonal"``.
        entangling_gate (str or :class:`qibo.gates.Gate`, optional): Two-qubit gate to be used
            in the entangling layer. If ``entangling_gate`` is a parametrized gate,
            all phases are initialized as :math:`0.0`. Defaults to  ``"CNOT"``.
        closed_boundary (bool, optional): If ``True`` adds a closed-boundary condition
            to the entangling layer. Defaults to ``False``.

    Returns:
        :class:`qibo.models.circuit.Circuit`: Circuit containing layer of two-qubit gates.
    """

    if not isinstance(nqubits, int):
        raise_error(
            TypeError, f"nqubits must be type int, but it is type {type(nqubits)}."
        )

    if nqubits <= 0.0:
        raise_error(
            ValueError, f"nqubits must be a positive integer, but it is {nqubits}."
        )

    if not isinstance(architecture, str):
        raise_error(
            TypeError,
            f"``architecture`` must be type str, but it is type {type(architecture)}.",
        )

    if architecture not in ["diagonal", "shifted", "even-layer", "odd-layer"]:
        raise_error(
            NotImplementedError,
            f"``architecture`` {architecture} not found.",
        )

    if not isinstance(closed_boundary, bool):
        raise_error(
            TypeError,
            f"closed_boundary must be type bool, but it is type {type(closed_boundary)}.",
        )

    gate = (
        getattr(gates, entangling_gate)
        if isinstance(entangling_gate, str)
        else entangling_gate
    )

    if gate.__name__ == "GeneralizedfSim":
        raise_error(
            NotImplementedError,
            "This function does not support the ``GeneralizedfSim`` gate.",
        )

    # Finds the number of correct number of parameters to initialize the gate class.
    parameters = list(signature(gate).parameters)

    if "q2" in parameters:
        raise_error(
            NotImplementedError, f"This function does not accept three-qubit gates."
        )

    # If gate is parametrized, sets all angles to 0.0
    parameters = (0.0,) * (len(parameters) - 3) if len(parameters) > 2 else None

    circuit = Circuit(nqubits)

    if architecture == "diagonal":
        circuit.add(
            _parametrized_two_qubit_gate(gate, qubit, qubit + 1, parameters)
            for qubit in range(nqubits - 1)
        )
    elif architecture == "even-layer":
        circuit.add(
            _parametrized_two_qubit_gate(gate, qubit, qubit + 1, parameters)
            for qubit in range(0, nqubits - 1, 2)
        )
    elif architecture == "odd-layer":
        circuit.add(
            _parametrized_two_qubit_gate(gate, qubit, qubit + 1, parameters)
            for qubit in range(1, nqubits - 1, 2)
        )
    else:
        circuit.add(
            _parametrized_two_qubit_gate(gate, qubit, qubit + 1, parameters)
            for qubit in range(0, nqubits - 1, 2)
        )
        circuit.add(
            _parametrized_two_qubit_gate(gate, qubit, qubit + 1, parameters)
            for qubit in range(1, nqubits - 1, 2)
        )

    if closed_boundary:
        circuit.add(_parametrized_two_qubit_gate(gate, nqubits - 1, 0, parameters))

    return circuit


def _generate_rbs_pairs(nqubits: int, architecture: str):
    """Generating list of indexes representing the RBS connections

    Creates circuit with all RBS initialised with 0.0 phase.

    Args:
        nqubits (int): number of qubits.
        architecture(str, optional): circuit architecture used for the unary loader.
            If ``diagonal``, uses a ladder-like structure.
            If ``tree``, uses a binary-tree-based structure.
            Defaults to ``tree``.

    Returns:
        (:class:`qibo.models.circuit.Circuit`, list): circuit composed of :class:`qibo.gates.gates.RBS`
            and list of indexes of target qubits per depth.
    """

    if architecture == "diagonal":
        pairs_rbs = np.arange(nqubits)
        pairs_rbs = [[pair] for pair in zip(pairs_rbs[:-1], pairs_rbs[1:])]

    if architecture == "tree":
        pairs_rbs = [[(0, int(nqubits / 2))]]
        indexes = list(pairs_rbs[0][0])
        for depth in range(2, int(math.log2(nqubits)) + 1):
            pairs_rbs_per_depth = [
                [(index, index + int(nqubits / 2**depth)) for index in indexes]
            ]
            pairs_rbs += pairs_rbs_per_depth
            indexes = list(np.array(pairs_rbs_per_depth).flatten())

    pairs_rbs = [
        [(nqubits - 1 - a, nqubits - 1 - b) for a, b in row] for row in pairs_rbs
    ]

    circuit = Circuit(nqubits)
    for row in pairs_rbs:
        for pair in row:
            circuit.add(gates.RBS(*pair, 0.0, trainable=True))

    return circuit, pairs_rbs


def _generate_rbs_angles(data, nqubits: int, architecture: str):
    """Generating list of angles for RBS gates based on ``architecture``.

    Args:
        data (ndarray, optional): :math:`1`-dimensional array of data to be loaded.
        nqubits (int): number of qubits.
        architecture(str, optional): circuit architecture used for the unary loader.
            If ``diagonal``, uses a ladder-like structure.
            If ``tree``, uses a binary-tree-based structure.
            Defaults to ``tree``.

    Returns:
        list: list of phases for RBS gates.
    """
    if architecture == "diagonal":
        phases = [
            math.atan2(np.linalg.norm(data[k + 1 :]), data[k])
            for k in range(len(data) - 2)
        ]
        phases.append(math.atan2(data[-1], data[-2]))

    if architecture == "tree":
        j_max = int(nqubits / 2)

        r_array = np.zeros(nqubits - 1, dtype=float)
        phases = np.zeros(nqubits - 1, dtype=float)
        for j in range(1, j_max + 1):
            r_array[j_max + j - 2] = math.sqrt(
                data[2 * j - 1] ** 2 + data[2 * j - 2] ** 2
            )
            theta = math.acos(data[2 * j - 2] / r_array[j_max + j - 2])
            if data[2 * j - 1] < 0.0:
                theta = 2 * math.pi - theta
            phases[j_max + j - 2] = theta

        for j in range(j_max - 1, 0, -1):
            r_array[j - 1] = math.sqrt(r_array[2 * j] ** 2 + r_array[2 * j - 1] ** 2)
            phases[j - 1] = math.acos(r_array[2 * j - 1] / r_array[j - 1])

    return phases


def _parametrized_two_qubit_gate(gate, q0, q1, params=None):
    """Returns two-qubit gate initialized with or without phases."""
    if params is not None:
        return gate(q0, q1, *params)

    return gate(q0, q1)



"""Qibo wrapper for QASM 3.0 parser."""

from itertools import repeat
from typing import Union

import numpy as np
import openqasm3

import qibo
from qibo.config import raise_error
from qibo.gates import FusedGate


class CustomQASMGate:
    """Object that handles the definition of custom gates in QASM via the `gate` command.

    Args:
        gates (list): List of gates composing the defined gate.
        qubits (list or tuple): Qubits identifiers (e.g. (q0, q1, q2, ...)).
        args (list or tuple): Arguments identifiers (e.g. (theta, alpha, gamma, ...)).
    """

    def __init__(
        self,
        name: str,
        gates: list,
        qubits: Union[list, tuple],
        args: Union[list, tuple],
    ):
        self.name = name
        self.gates = gates
        self.qubits = qubits
        self.args = args

    def get_gate(self, qubits: Union[list, tuple], args: Union[list, tuple]):
        """Returns the gates composing the defined gate applied on the
        specified qubits with the specified ``args`` as a unique :class:`qibo.gates.special.FusedGate`.

        Args:
            qubits (list or tuple): Qubits where to apply the gates.
            args (list or tuple): Arguments to evaluate the gates on.

        Returns:
            :class:`qibo.gates.special.FusedGate`: the composed gate evaluated on the input qubits with the input arguments.
        """
        if len(self.args) != len(args):
            raise_error(
                ValueError,
                f"Invalid `args` argument passed to the user-defined gate `{self.name}` upon construction. {args} was passed but something of the form {self.args} is expected.",
            )
        elif len(self.qubits) != len(qubits):
            raise_error(
                ValueError,
                f"Invalid `qubits` argument passed to the user-defined gate `{self.name}` upon construction. {qubits} was passed but something of the form {self.qubits} is expected.",
            )
        qubit_map = dict(zip(self.qubits, qubits))
        args_map = dict(zip(self.args, args))
        return self._construct_fused_gate(self.gates, qubits, qubit_map, args_map)

    def _construct_fused_gate(self, gates, qubits, qubit_map, args_map):
        """Constructs a :class:`qibo.gates.special.FusedGate` out of the provided list of gates on the specified qubits.

        Args:
            gates (list(:class:`qibo.gates.Gate`)): List of gates to build the fused gate from.
            qubits (list(int)): List of qubits to construct the gate on.
            qubit_map (dict): Mapping between the placeholders for the qubits contained in `gates` and the actual qubits indices to apply them on.
            args_map (dict): Mapping between the placeholders for the kwargs contained in `gates` and the actual kwargs values.

        Returns:
            (qibo.gates.special.FusedGate): The resulting fused gate.
        """
        fused_gate = FusedGate(*qubits)
        for gate in gates:
            if not isinstance(gate, FusedGate):
                new_qubits, new_args = self._compile_gate_qubits_and_args(
                    gate, qubit_map, args_map
                )
                fused_gate.append(gate.__class__(*new_qubits, *new_args))
            else:
                qubits = [qubit_map[q] for q in gate.qubits]
                fused_gate.append(
                    self._construct_fused_gate(gate.gates, qubits, qubit_map, args_map)
                )
        return fused_gate

    def _compile_gate_qubits_and_args(self, gate, qubit_map, args_map):
        """Compile the qubits and arguments placeholders contained in the input gate with their actual values.

        Args:
            gate (:class:`qibo.gates.Gate`): The input gate containing the qubits and arguments placeholders.
            qubit_map (dict): Mapping between the placeholders for the qubits contained in `gate` and the actual qubits indices to apply them on.
            args_map (dict): Mapping between the placeholders for the kwargs contained in `gate` and the actual kwargs values.

        Returns:
            tuple(list, list): The compiled qubits and arguments.
        """
        new_qubits = [qubit_map[q] for q in gate.qubits]
        new_args = [args_map.get(arg, arg) for arg in gate.init_kwargs.values()]
        return new_qubits, new_args


def _qibo_gate_name(gate):
    if gate == "cx":
        return "CNOT"

    if gate == "id":
        return "I"

    if gate == "ccx":
        return "TOFFOLI"

    return gate.upper()


class QASMParser:
    """Wrapper around the :class:`openqasm3.parser` for QASM 3.0."""

    def __init__(
        self,
    ):
        self.parser = openqasm3.parser
        self.defined_gates = {}
        self.q_registers = {}
        self.c_registers = set()

    def to_circuit(
        self, qasm_string: str, accelerators: dict = None, density_matrix: bool = False
    ):
        """Converts a QASM program into a :class:`qibo.models.Circuit`.

        Args:
            qasm_string (str): QASM program.
            accelerators (dict, optional): Maps device names to the number of times each
                device will be used. Defaults to ``None``.
            density_matrix (bool, optional): If ``True``, the constructed circuit would
                evolve density matrices.

        Returns:
            :class:`qibo.models.Circuit`: circuit constructed from QASM string.
        """
        parsed = self.parser.parse(qasm_string)
        gates = []
        self.defined_gates, self.q_registers, self.c_registers = {}, {}, {}

        nqubits = 0
        for statement in parsed.statements:
            if isinstance(statement, openqasm3.ast.QuantumGate):
                gates.append(self._get_gate(statement))
            elif isinstance(statement, openqasm3.ast.QuantumMeasurementStatement):
                gates.append(self._get_measurement(statement))
            elif isinstance(statement, openqasm3.ast.QubitDeclaration):
                q_name, q_size = self._get_qubit(statement)
                self.q_registers.update(
                    {q_name: list(range(nqubits, nqubits + q_size))}
                )
                nqubits += q_size
            elif isinstance(statement, openqasm3.ast.QuantumGateDefinition):
                self._def_gate(statement)
            elif isinstance(statement, openqasm3.ast.Include):
                continue
            elif isinstance(statement, openqasm3.ast.ClassicalDeclaration):
                name = statement.identifier.name
                size = statement.type.size.value
                self.c_registers.update({name: list(range(size))})
            else:
                raise_error(RuntimeError, f"Unsupported {type(statement)} statement.")
        circ = qibo.Circuit(
            nqubits,
            accelerators,
            density_matrix,
            wire_names=self._construct_wire_names(),
        )
        circ.add(self._merge_measurements(gates))
        return circ

    def _get_measurement(self, measurement):
        """Converts a :class:`openqasm3.ast.QuantumMeasurementStatement` statement
        into :class:`qibo.gates.measurements.M`."""
        qubit = self._get_qubit(measurement.measure.qubit)
        register = measurement.target.name.name
        if register not in self.c_registers:
            raise_error(ValueError, f"Undefined measurement register `{register}`.")
        ind = measurement.target.indices[0][0].value
        if ind >= len(self.c_registers[register]):
            raise_error(
                IndexError, f"Index `{ind}` is out of bounds of register `{register}`."
            )
        self.c_registers[register][ind] = qubit
        return getattr(qibo.gates, "M")(qubit, register_name=register)

    def _get_qubit(self, qubit):
        """Extracts the qubit from a :class:`openqasm3.ast.QubitDeclaration` statement."""
        if isinstance(qubit, openqasm3.ast.QubitDeclaration):
            return qubit.qubit.name, qubit.size.value

        if not isinstance(qubit, openqasm3.ast.Identifier):
            return self.q_registers[qubit.name.name][qubit.indices[0][0].value]

        return qubit.name

    def _get_gate(self, gate):
        """Converts a :class:`openqasm3.ast.QuantumGate` statement
        into :class:`qibo.gates.Gate`."""
        qubits = [self._get_qubit(q) for q in gate.qubits]
        init_args = []
        for arg in gate.arguments:
            arg = self._unroll_expression(arg)
            try:
                arg = eval(arg.replace("pi", "np.pi"))
            except:
                pass
            init_args.append(arg)
        # check whether the gate exists in qibo.gates already
        if _qibo_gate_name(gate.name.name) in dir(qibo.gates):
            try:
                gate = getattr(qibo.gates, _qibo_gate_name(gate.name.name))(
                    *qubits, *init_args
                )
            # the gate exists in qibo.gates but invalid construction
            except TypeError:
                raise_error(
                    ValueError, f"Invalid gate declaration at span: {gate.span}"
                )
        # check whether the gate was defined by the user
        elif gate.name.name in self.defined_gates:
            try:
                gate = self.defined_gates.get(gate.name.name).get_gate(
                    qubits, init_args
                )
            # the gate exists in self.defined_gates but invalid construction
            except ValueError:
                raise_error(
                    ValueError, f"Invalid gate declaration at span: {gate.span}"
                )
        # undefined gate
        else:
            raise_error(ValueError, f"Undefined gate at span: {gate.span}")
        return gate

    def _unroll_expression(self, expr):
        """Unrolls an argument definition expression to retrieve the
        complete argument as a string."""
        # check whether the expression is a simple string, e.g. `pi` or `theta`
        if "name" in dir(expr):
            return expr.name
        # check whether the expression is a single value, e.g. `0.1234`
        if "value" in dir(expr):
            return expr.value
        # the expression is composite, e.g. `2*pi` or `3*theta/2`
        expr_dict = {}
        for attr in ("lhs", "op", "expression", "rhs"):
            expr_dict[attr] = ""
            if attr in dir(expr):
                val = self._unroll_expression(getattr(expr, attr))
                expr_dict[attr] += str(val)

        return "".join(list(expr_dict.values()))

    def _def_gate(self, definition):
        """Converts a :class:`openqasm3.ast.QuantumGateDefinition` statement
        into :class:`qibo.parser.CustomQASMGate` object."""
        name = definition.name.name
        qubits = [self._get_qubit(q) for q in definition.qubits]
        args = [self._unroll_expression(expr) for expr in definition.arguments]
        gates = [self._get_gate(gate) for gate in definition.body]
        self.defined_gates.update({name: CustomQASMGate(name, gates, qubits, args)})

    def _merge_measurements(self, gates):
        """Merges separated measurements of a same register into a single one.
        This is needed because qibo doesn't allow to separetely define two measurements in a same register:

        # not allowed
        c.add(gates.M(0, register="m0"))
        c.add(gates.M(1, register="m0"))
        """
        updated_queue = []
        for gate in gates:
            if isinstance(gate, qibo.gates.M):
                if gate.register_name in self.c_registers:
                    updated_queue.append(
                        qibo.gates.M(
                            *self.c_registers.pop(gate.register_name),
                            register_name=gate.register_name,
                        )
                    )
            else:
                updated_queue.append(gate)
        return updated_queue

    def _construct_wire_names(self):
        """Builds the wires names from the declared quantum registers."""
        wire_names = []
        for reg_name, reg_qubits in self.q_registers.items():
            wires = sorted(
                zip(repeat(reg_name, len(reg_qubits)), reg_qubits), key=lambda x: x[1]
            )
            for wire in wires:
                wire_names.append(f"{wire[0]}{wire[1]}")
        return wire_names



from abc import abstractmethod

from qibo.config import raise_error


class AbstractHamiltonian:
    """Qibo abstraction for Hamiltonian objects."""

    def __init__(self):
        self._nqubits = None

    @property
    def nqubits(self):
        return self._nqubits

    @nqubits.setter
    def nqubits(self, n):
        if not isinstance(n, int):
            raise_error(RuntimeError, f"nqubits must be an integer but is {type(n)}.")
        if n < 1:
            raise_error(ValueError, f"nqubits must be a positive integer but is {n}")
        self._nqubits = n

    @abstractmethod
    def eigenvalues(self, k=6):  # pragma: no cover
        """Computes the eigenvalues for the Hamiltonian.

        Args:
            k (int): Number of eigenvalues to calculate if the Hamiltonian
                was created using a sparse matrix. This argument is ignored
                if the Hamiltonian was created using a dense matrix.
                See :meth:`qibo.backends.abstract.AbstractBackend.eigvalsh` for
                more details.
        """
        raise_error(NotImplementedError)

    @abstractmethod
    def eigenvectors(self, k=6):  # pragma: no cover
        """Computes a tensor with the eigenvectors for the Hamiltonian.

        Args:
            k (int): Number of eigenvalues to calculate if the Hamiltonian
                was created using a sparse matrix. This argument is ignored
                if the Hamiltonian was created using a dense matrix.
                See :meth:`qibo.backends.abstract.AbstractBackend.eigh` for
                more details.
        """
        raise_error(NotImplementedError)

    def ground_state(self):
        """Computes the ground state of the Hamiltonian.

        Uses :meth:`qibo.hamiltonians.AbstractHamiltonian.eigenvectors`
        and returns eigenvector corresponding to the lowest energy.
        """
        return self.eigenvectors()[:, 0]

    @abstractmethod
    def exp(self, a):  # pragma: no cover
        """Computes a tensor corresponding to exp(-1j * a * H).

        Args:
            a (complex): Complex number to multiply Hamiltonian before
                exponentiation.
        """
        raise_error(NotImplementedError)

    @abstractmethod
    def expectation(self, state, normalize=False):  # pragma: no cover
        """Computes the real expectation value for a given state.

        Args:
            state (array): the expectation state.
            normalize (bool): If ``True`` the expectation value is divided
                with the state's norm squared.

        Returns:
            Real number corresponding to the expectation value.
        """
        raise_error(NotImplementedError)

    @abstractmethod
    def expectation_from_samples(self, freq, qubit_map=None):  # pragma: no cover
        """Computes the real expectation value of a diagonal observable given the frequencies when measuring in the computational basis.

        Args:
            freq (collections.Counter): the keys are the observed values in binary form
            and the values the corresponding frequencies, that is the number
            of times each measured value/bitstring appears.
            qubit_map (tuple): Mapping between frequencies and qubits. If None, [1,...,len(key)]

        Returns:
            Real number corresponding to the expectation value.
        """
        raise_error(NotImplementedError)

    @abstractmethod
    def __add__(self, o):  # pragma: no cover
        """Add operator."""
        raise_error(NotImplementedError)

    def __radd__(self, o):
        """Right operator addition."""
        return self.__add__(o)

    @abstractmethod
    def __sub__(self, o):  # pragma: no cover
        """Subtraction operator."""
        raise_error(NotImplementedError)

    @abstractmethod
    def __rsub__(self, o):  # pragma: no cover
        """Right subtraction operator."""
        raise_error(NotImplementedError)

    @abstractmethod
    def __mul__(self, o):  # pragma: no cover
        """Multiplication to scalar operator."""
        raise_error(NotImplementedError)

    def __rmul__(self, o):
        """Right scalar multiplication."""
        return self.__mul__(o)

    @abstractmethod
    def __matmul__(self, o):  # pragma: no cover
        """Matrix multiplication with other Hamiltonians or state vectors."""
        raise_error(NotImplementedError)



import collections
import json
from typing import List, Sequence, Tuple

import sympy

from qibo.backends import _check_backend
from qibo.config import raise_error

REQUIRED_FIELDS = [
    "name",
    "init_args",
    "init_kwargs",
    "_target_qubits",
    "_control_qubits",
]
REQUIRED_FIELDS_INIT_KWARGS = ["theta", "phi", "lam", "phi0", "phi1"]


class Gate:
    """The base class for gate implementation.

    All base gates should inherit this class.

    Attributes:
        name (str): Name of the gate.
        draw_label (str): Optional label for drawing the gate in a circuit
            with :func:`qibo.models.Circuit.draw`.
        is_controlled_by (bool): ``True`` if the gate was created using the
            :meth:`qibo.gates.abstract.Gate.controlled_by` method,
            otherwise ``False``.
        init_args (list): Arguments used to initialize the gate.
        init_kwargs (dict): Arguments used to initialize the gate.
        target_qubits (tuple): Tuple with ids of target qubits.
        control_qubits (tuple): Tuple with ids of control qubits sorted in
            increasing order.
    """

    def __init__(self):
        from qibo import config

        self.name = None
        self.draw_label = None
        self.is_controlled_by = False
        # args for creating gate
        self.init_args = []
        self.init_kwargs = {}

        self.unitary = False
        self._target_qubits = ()
        self._control_qubits = ()
        self._parameters = ()
        config.ALLOW_SWITCHERS = False

        self.symbolic_parameters = {}

        # for distributed circuits
        self.device_gates = set()
        self.original_gate = None

    @property
    def clifford(self):
        """Return boolean value representing if a Gate is Clifford or not."""
        return False

    @property
    def raw(self) -> dict:
        """Serialize to dictionary.

        The values used in the serialization should be compatible with a
        JSON dump (or any other one supporting a minimal set of scalar
        types). Though the specific implementation is up to the specific
        gate.
        """
        encoded = self.__dict__

        encoded_simple = {
            key: value for key, value in encoded.items() if key in REQUIRED_FIELDS
        }

        encoded_simple["init_kwargs"] = {
            key: value
            for key, value in encoded_simple["init_kwargs"].items()
            if key in REQUIRED_FIELDS_INIT_KWARGS
        }

        for value in encoded_simple:
            if isinstance(encoded[value], set):
                encoded_simple[value] = list(encoded_simple[value])

        encoded_simple["_class"] = type(self).__name__

        return encoded_simple

    @staticmethod
    def from_dict(raw: dict):
        """Load from serialization.

        Essentially the counter-part of :meth:`raw`.
        """
        from qibo.gates import gates, measurements

        for mod in (gates, measurements):
            try:
                cls = getattr(mod, raw["_class"])
                break
            except AttributeError:
                # gate not found in given module, try next
                pass
        else:
            raise ValueError(f"Unknown gate {raw['_class']}")

        gate = cls(*raw["init_args"], **raw["init_kwargs"])
        try:
            return gate.controlled_by(*raw["_control_qubits"])
        except RuntimeError as e:
            if "controlled" in e.args[0]:
                return gate
            raise e

    def to_json(self):
        """Dump gate to JSON.

        Note:
            Consider using :meth:`raw` directly.
        """
        return json.dumps(self.raw)

    @property
    def target_qubits(self) -> Tuple[int]:
        """Tuple with ids of target qubits."""
        return self._target_qubits

    @property
    def control_qubits(self) -> Tuple[int]:
        """Tuple with ids of control qubits sorted in increasing order."""
        return tuple(sorted(self._control_qubits))

    @property
    def qubits(self) -> Tuple[int]:
        """Tuple with ids of all qubits (control and target) that the gate acts."""
        return self.control_qubits + self.target_qubits

    @property
    def qasm_label(self):
        """String corresponding to OpenQASM operation of the gate."""
        raise_error(
            NotImplementedError,
            f"{self.__class__.__name__} is not supported by OpenQASM",
        )

    def _set_target_qubits(self, qubits: Sequence[int]):
        """Helper method for setting target qubits."""
        self._target_qubits = tuple(qubits)
        if len(self._target_qubits) != len(set(qubits)):
            repeated = self._find_repeated(qubits)
            raise_error(
                ValueError,
                f"Target qubit {repeated} was given twice for gate {self.__class__.__name__}.",
            )

    def _set_control_qubits(self, qubits: Sequence[int]):
        """Helper method for setting control qubits."""
        if len(set(qubits)) != len(qubits):
            repeated = self._find_repeated(qubits)
            raise_error(
                ValueError,
                f"Control qubit {repeated} was given twice for gate {self.__class__.__name__}.",
            )
        self._control_qubits = qubits

    @target_qubits.setter
    def target_qubits(self, qubits: Sequence[int]):
        """Sets target qubits tuple."""
        self._set_target_qubits(qubits)
        self._check_control_target_overlap()

    @control_qubits.setter
    def control_qubits(self, qubits: Sequence[int]):
        """Sets control qubits set."""
        self._set_control_qubits(qubits)
        self._check_control_target_overlap()

    def _set_targets_and_controls(
        self, target_qubits: Sequence[int], control_qubits: Sequence[int]
    ):
        """Sets target and control qubits simultaneously.

        This is used for the reduced qubit updates in the distributed
        circuits because using the individual setters may raise errors
        due to temporary overlap of control and target qubits.
        """
        self._set_target_qubits(target_qubits)
        self._set_control_qubits(control_qubits)
        self._check_control_target_overlap()

    @staticmethod
    def _find_repeated(qubits: Sequence[int]) -> int:
        """Finds the first qubit id that is repeated in a sequence of qubit ids."""
        temp_set = set()
        for qubit in qubits:
            if qubit in temp_set:
                return qubit
            temp_set.add(qubit)

    def _check_control_target_overlap(self):
        """Checks that there are no qubits that are both target and
        controls."""
        control_and_target = self._control_qubits + self._target_qubits
        common = len(set(control_and_target)) != len(control_and_target)
        if common:
            raise_error(
                ValueError,
                f"{set(self._target_qubits) & set(self._control_qubits)}"
                + "qubits are both targets and controls "
                + f"for gate {self.__class__.__name__}.",
            )

    @property
    def parameters(self):
        """Returns a tuple containing the current value of gate's parameters."""
        return self._parameters

    def commutes(self, gate: "Gate") -> bool:
        """Checks if two gates commute.

        Args:
            gate: Gate to check if it commutes with the current gate.

        Returns:
            bool: ``True`` if the gates commute, ``False`` otherwise.
        """
        if isinstance(gate, SpecialGate):  # pragma: no cover
            return False
        t1 = set(self.target_qubits)
        t2 = set(gate.target_qubits)
        a = self.__class__ == gate.__class__ and t1 == t2
        b = not (t1 & set(gate.qubits) or t2 & set(self.qubits))
        return a or b

    def on_qubits(self, qubit_map) -> "Gate":
        """Creates the same gate targeting different qubits.

        Args:
            qubit_map (int): Dictionary mapping original qubit indices to new ones.

        Returns:
            A :class:`qibo.gates.Gate` object of the original gate
            type targeting the given qubits.

        Example:

            .. testcode::

                from qibo import models, gates
                c = models.Circuit(4)
                # Add some CNOT gates
                c.add(gates.CNOT(2, 3).on_qubits({2: 2, 3: 3})) # equivalent to gates.CNOT(2, 3)
                c.add(gates.CNOT(2, 3).on_qubits({2: 3, 3: 0})) # equivalent to gates.CNOT(3, 0)
                c.add(gates.CNOT(2, 3).on_qubits({2: 1, 3: 3})) # equivalent to gates.CNOT(1, 3)
                c.add(gates.CNOT(2, 3).on_qubits({2: 2, 3: 1})) # equivalent to gates.CNOT(2, 1)
                print(c.draw())
            .. testoutput::

                q0: ───X─────
                q1: ───|─o─X─
                q2: ─o─|─|─o─
                q3: ─X─o─X───
        """
        if self.is_controlled_by:
            targets = (qubit_map.get(q) for q in self.target_qubits)
            controls = (qubit_map.get(q) for q in self.control_qubits)
            gate = self.__class__(*targets, **self.init_kwargs)
            gate = gate.controlled_by(*controls)
        else:
            qubits = (qubit_map.get(q) for q in self.qubits)
            gate = self.__class__(*qubits, **self.init_kwargs)
        return gate

    def _dagger(self) -> "Gate":
        """Helper method for :meth:`qibo.gates.Gate.dagger`."""
        # By default the ``_dagger`` method creates an equivalent gate, assuming
        # that the gate is Hermitian (true for common gates like H or Paulis).
        # If the gate is not Hermitian the ``_dagger`` method should be modified.
        return self.__class__(*self.init_args, **self.init_kwargs)

    def dagger(self) -> "Gate":
        """Returns the dagger (conjugate transpose) of the gate.

        Note that dagger is not persistent for parametrized gates.
        For example, applying a dagger to an :class:`qibo.gates.gates.RX` gate
        will change the sign of its parameter at the time of application.
        However, if the parameter is updated after that, for example using
        :meth:`qibo.models.circuit.Circuit.set_parameters`, then the
        action of dagger will be lost.

        Returns:
            :class:`qibo.gates.Gate`: object representing the dagger of the original gate.
        """
        new_gate = self._dagger()
        new_gate.is_controlled_by = self.is_controlled_by
        new_gate.control_qubits = self.control_qubits
        return new_gate

    def check_controls(func):  # pylint: disable=E0213
        def wrapper(self, *args):
            if self.control_qubits:
                raise_error(
                    RuntimeError,
                    "Cannot use `controlled_by` method "
                    + f"on gate {self} because it is already "
                    + f"controlled by {self.control_qubits}.",
                )
            return func(self, *args)  # pylint: disable=E1102

        return wrapper

    @check_controls
    def controlled_by(self, *qubits: int) -> "Gate":
        """Controls the gate on (arbitrarily many) qubits.

        To see how this method affects the underlying matrix representation of a gate,
        please see the documentation of :meth:`qibo.gates.Gate.matrix`.

        .. note::
            Some gate classes default to another gate class depending on the number of controls
            present. For instance, an :math:`1`-controlled :class:`qibo.gates.X` gate
            will default to a :class:`qibo.gates.CNOT` gate, while a :math:`2`-controlled
            :class:`qibo.gates.X` gate defaults to a :class:`qibo.gates.TOFFOLI` gate.
            Other gates affected by this method are: :class:`qibo.gates.Y`, :class:`qibo.gates.Z`,
            :class:`qibo.gates.RX`, :class:`qibo.gates.RY`, :class:`qibo.gates.RZ`,
            :class:`qibo.gates.U1`, :class:`qibo.gates.U2`, and :class:`qibo.gates.U3`.

        Args:
            *qubits (int): Ids of the qubits that the gate will be controlled on.

        Returns:
            :class:`qibo.gates.Gate`: object in with the corresponding
                gate being controlled in the given qubits.
        """
        if qubits:
            self.is_controlled_by = True
            self.control_qubits = qubits
        return self

    def decompose(self, *free) -> List["Gate"]:
        """Decomposes multi-control gates to gates supported by OpenQASM.

        Decompositions are based on `arXiv:9503016 <https://arxiv.org/abs/quant-ph/9503016>`_.

        Args:
            free: Ids of free qubits to use for the gate decomposition.

        Returns:
            list: gates that have the same effect as applying the original gate.
        """
        # TODO: Implement this method for all gates not supported by OpenQASM.
        # Currently this is implemented only for multi-controlled X gates.
        # If it is used on a different gate it will just return a deep copy
        # of the same gate.
        return [self.__class__(*self.init_args, **self.init_kwargs)]

    def matrix(self, backend=None):
        """Returns the matrix representation of the gate.

        If gate has controlled qubits inserted by :meth:`qibo.gates.Gate.controlled_by`,
        then :meth:`qibo.gates.Gate.matrix` returns the matrix of the original gate.

        .. code-block:: python

            from qibo import gates

            gate = gates.SWAP(3, 4).controlled_by(0, 1, 2)
            print(gate.matrix())

        To return the full matrix that takes the control qubits into account,
        one should use :meth:`qibo.models.Circuit.unitary`, e.g.

        .. code-block:: python

            from qibo import Circuit, gates

            nqubits = 5
            circuit = Circuit(nqubits)
            circuit.add(gates.SWAP(3, 4).controlled_by(0, 1, 2))
            print(circuit.unitary())

        Args:
            backend (:class:`qibo.backends.abstract.Backend`, optional): backend
                to be used in the execution. If ``None``, it uses
                :class:`qibo.backends.GlobalBackend`. Defaults to ``None``.

        Returns:
            ndarray: Matrix representation of gate.

        .. note::
            ``Gate.matrix`` was defined as an atribute in ``qibo`` versions prior to  ``0.2.0``.
            From ``0.2.0`` on, it has been converted into a method and has replaced the ``asmatrix`` method.
        """
        backend = _check_backend(backend)

        return backend.matrix(self)

    def generator_eigenvalue(self):
        """This function returns the eigenvalues of the gate's generator.

        Returns:
            float: eigenvalue of the generator.
        """

        raise_error(
            NotImplementedError,
            f"Generator eigenvalue is not implemented for {self.__class__.__name__}",
        )

    def basis_rotation(self):
        """Transformation required to rotate the basis for measuring the gate."""
        raise_error(
            NotImplementedError,
            f"Basis rotation is not implemented for {self.__class__.__name__}",
        )

    def apply(self, backend, state, nqubits):
        return backend.apply_gate(self, state, nqubits)

    def apply_density_matrix(self, backend, state, nqubits):
        return backend.apply_gate_density_matrix(self, state, nqubits)

    def apply_clifford(self, backend, state, nqubits):
        return backend.apply_gate_clifford(self, state, nqubits)


class SpecialGate(Gate):
    """Abstract class for special gates."""

    def commutes(self, gate):
        return False

    def on_qubits(self, qubit_map):
        raise_error(NotImplementedError, "Cannot use special gates on subroutines.")

    def matrix(self, backend=None):  # pragma: no cover
        raise_error(
            NotImplementedError, "Special gates do not have matrix representation."
        )


class ParametrizedGate(Gate):
    """Base class for parametrized gates.

    Implements the basic functionality of parameter setters and getters.
    """

    def __init__(self, trainable=True):
        super().__init__()
        self.parameter_names = "theta"
        self.nparams = 1
        self.trainable = trainable

    @Gate.parameters.setter
    def parameters(self, x):
        """Updates the values of gate's parameters."""
        if isinstance(self.parameter_names, str):
            nparams = 1
            names = [self.parameter_names]
            if not isinstance(x, collections.abc.Iterable):
                x = [x]
            else:
                # Captures the ``Unitary`` gate case where the given parameter
                # can be an array
                try:
                    if len(x) != 1:  # pragma: no cover
                        x = [x]
                except TypeError:  # tf.Variable case
                    s = tuple(x.shape)
                    if not s or s[0] != 1:
                        x = [x]
        else:
            nparams = len(self.parameter_names)
            names = self.parameter_names

        if not self._parameters:
            params = nparams * [None]
        else:
            params = list(self._parameters)
        if len(x) != nparams:
            raise_error(
                ValueError,
                f"Parametrized gate has {nparams} parameters "
                + f"but {len(x)} update values were given.",
            )
        for i, v in enumerate(x):
            if isinstance(v, sympy.Expr):
                self.symbolic_parameters[i] = v
            params[i] = v
        self._parameters = tuple(params)
        self.init_kwargs.update(
            {n: v for n, v in zip(names, self._parameters) if n in self.init_kwargs}
        )

        # set parameters in device gates
        for gate in self.device_gates:  # pragma: no cover
            gate.parameters = x

    def on_qubits(self, qubit_map):
        gate = super().on_qubits(qubit_map)
        gate.parameters = self.parameters
        return gate

    def substitute_symbols(self):
        params = list(self._parameters)
        for i, param in self.symbolic_parameters.items():
            for symbol in param.free_symbols:
                param = symbol.evaluate(param)
            params[i] = float(param)
        self.parameters = tuple(params)

    def matrix(self, backend=None):
        backend = _check_backend(backend)

        return backend.matrix_parametrized(self)



from qibo.quantum_info.basis import *
from qibo.quantum_info.clifford import *
from qibo.quantum_info.entanglement import *
from qibo.quantum_info.entropies import *
from qibo.quantum_info.metrics import *
from qibo.quantum_info.operations import *
from qibo.quantum_info.quantum_networks import *
from qibo.quantum_info.random_ensembles import *
from qibo.quantum_info.superoperator_transformations import *
from qibo.quantum_info.utils import *



from typing import Optional

import networkx as nx
import numpy as np

from qibo.backends import NumpyBackend
from qibo.config import raise_error
from qibo.models import Circuit
from qibo.quantum_info.random_ensembles import random_statevector
from qibo.transpiler._exceptions import TranspilerPipelineError
from qibo.transpiler.abstract import Optimizer, Placer, Router
from qibo.transpiler.optimizer import Preprocessing
from qibo.transpiler.placer import StarConnectivityPlacer, Trivial, assert_placement
from qibo.transpiler.router import (
    ConnectivityError,
    StarConnectivityRouter,
    assert_connectivity,
)
from qibo.transpiler.unroller import (
    DecompositionError,
    NativeGates,
    Unroller,
    assert_decomposition,
)


def assert_circuit_equivalence(
    original_circuit: Circuit,
    transpiled_circuit: Circuit,
    final_map: dict,
    initial_map: Optional[dict] = None,
    test_states: Optional[list] = None,
    ntests: int = 3,
):
    """Checks that the transpiled circuit agrees with the original using simulation.

    Args:
        original_circuit (:class:`qibo.models.circuit.Circuit`): Original circuit.
        transpiled_circuit (:class:`qibo.models.circuit.Circuit`): Transpiled circuit.
        final_map (dict): logical-physical qubit mapping after routing.
        initial_map (dict, optional): logical_physical qubit mapping before routing.
            If ``None``, trivial initial map is used. Defauts to ``None``.
        test_states (list, optional): states on which the test is performed.
            If ``None``, ``ntests`` random states will be tested. Defauts to ``None``.
        ntests (int, optional): number of random states tested. Defauts to :math:`3`.
    """
    backend = NumpyBackend()
    ordering = np.argsort(np.array(list(final_map.values())))
    if transpiled_circuit.nqubits != original_circuit.nqubits:
        raise_error(
            ValueError,
            "Transpiled and original circuit do not have the same number of qubits.",
        )

    if test_states is None:
        test_states = [
            random_statevector(dims=2**original_circuit.nqubits, backend=backend)
            for _ in range(ntests)
        ]
    if initial_map is not None:
        reordered_test_states = []
        initial_map = np.array(list(initial_map.values()))
        reordered_test_states = [
            _transpose_qubits(initial_state, initial_map)
            for initial_state in test_states
        ]
    else:
        reordered_test_states = test_states

    for i in range(len(test_states)):
        target_state = backend.execute_circuit(
            original_circuit, initial_state=test_states[i]
        ).state()
        final_state = backend.execute_circuit(
            transpiled_circuit, initial_state=reordered_test_states[i]
        ).state()
        final_state = _transpose_qubits(final_state, ordering)
        fidelity = np.abs(np.dot(np.conj(target_state), final_state))
        try:
            np.testing.assert_allclose(fidelity, 1.0)
        except AssertionError:
            raise_error(TranspilerPipelineError, "Circuit equivalence not satisfied.")


def _transpose_qubits(state: np.ndarray, qubits_ordering: np.ndarray):
    """Reorders qubits of a given state vector.

    Args:
        state (np.ndarray): final state of the circuit.
        qubits_ordering (np.ndarray): final qubit ordering.
    """
    original_shape = state.shape
    state = np.reshape(state, len(qubits_ordering) * (2,))
    state = np.transpose(state, qubits_ordering)
    return np.reshape(state, original_shape)


def assert_transpiling(
    original_circuit: Circuit,
    transpiled_circuit: Circuit,
    connectivity: nx.Graph,
    initial_layout: dict,
    final_layout: dict,
    native_gates: NativeGates = NativeGates.default(),
    check_circuit_equivalence=True,
):
    """Check that all transpiler passes have been executed correctly.

    Args:
        original_circuit (qibo.models.Circuit): circuit before transpiling.
        transpiled_circuit (qibo.models.Circuit): circuit after transpiling.
        connectivity (networkx.Graph): chip qubits connectivity.
        initial_layout (dict): initial physical-logical qubit mapping.
        final_layout (dict): final physical-logical qubit mapping.
        native_gates (NativeGates): native gates supported by the hardware.
        check_circuit_equivalence (Bool): use simulations to check if the transpiled circuit is the same as the original.
    """
    assert_connectivity(circuit=transpiled_circuit, connectivity=connectivity)
    assert_decomposition(
        circuit=transpiled_circuit,
        native_gates=native_gates,
    )
    if original_circuit.nqubits != transpiled_circuit.nqubits:
        qubit_matcher = Preprocessing(connectivity=connectivity)
        original_circuit = qubit_matcher(circuit=original_circuit)
    assert_placement(
        circuit=original_circuit, layout=initial_layout, connectivity=connectivity
    )
    assert_placement(
        circuit=transpiled_circuit, layout=final_layout, connectivity=connectivity
    )
    if check_circuit_equivalence:
        assert_circuit_equivalence(
            original_circuit=original_circuit,
            transpiled_circuit=transpiled_circuit,
            initial_map=initial_layout,
            final_map=final_layout,
        )


def restrict_connectivity_qubits(connectivity: nx.Graph, qubits: list):
    """Restrict the connectivity to selected qubits.

    Args:
        connectivity (:class:`networkx.Graph`): chip connectivity.
        qubits (list): list of physical qubits to be used.

    Returns:
        (:class:`networkx.Graph`): restricted connectivity.
    """
    if not set(qubits).issubset(set(connectivity.nodes)):
        raise_error(
            ConnectivityError, "Some qubits are not in the original connectivity."
        )

    new_connectivity = nx.Graph()
    new_connectivity.add_nodes_from(qubits)
    new_edges = [
        edge for edge in connectivity.edges if edge[0] in qubits and edge[1] in qubits
    ]
    new_connectivity.add_edges_from(new_edges)

    if not nx.is_connected(new_connectivity):
        raise_error(ConnectivityError, "New connectivity graph is not connected.")

    return new_connectivity


class Passes:
    """Define a transpiler pipeline consisting of smaller transpiler steps that are applied sequentially:

    Args:
        passes (list, optional): list of passes to be applied sequentially.
            If ``None``, default transpiler will be used.
            Defaults to ``None``.
        connectivity (:class:`networkx.Graph`, optional): physical qubits connectivity.
            If ``None``, :class:`` is used.
            Defaults to ``None``.
        native_gates (:class:`qibo.transpiler.unroller.NativeGates`, optional): native gates.
            Defaults to :math:`qibo.transpiler.unroller.NativeGates.default`.
        on_qubits (list, optional): list of physical qubits to be used.
            If "None" all qubits are used. Defaults to ``None``.
        int_qubit_name (bool, optional): if `True` the `final_layout` keys are
            cast to integers.
    """

    def __init__(
        self,
        passes: list = None,
        connectivity: nx.Graph = None,
        native_gates: NativeGates = NativeGates.default(),
        on_qubits: list = None,
        int_qubit_names: bool = False,
    ):
        if on_qubits is not None:
            connectivity = restrict_connectivity_qubits(connectivity, on_qubits)
        self.connectivity = connectivity
        self.native_gates = native_gates
        self.passes = self.default() if passes is None else passes
        self.initial_layout = None
        self.int_qubit_names = int_qubit_names

    def default(self):
        """Return the default transpiler pipeline for the required hardware connectivity."""
        if not isinstance(self.connectivity, nx.Graph):
            raise_error(
                TranspilerPipelineError,
                "Define the hardware chip connectivity to use default transpiler",
            )
        default_passes = []
        # preprocessing
        default_passes.append(Preprocessing(connectivity=self.connectivity))
        # default placer pass
        default_passes.append(StarConnectivityPlacer())
        # default router pass
        default_passes.append(StarConnectivityRouter())
        # default unroller pass
        default_passes.append(Unroller(native_gates=self.native_gates))

        return default_passes

    def __call__(self, circuit):
        """
        This function returns the compiled circuits and the dictionary mapping
        physical (keys) to logical (values) qubit. If `int_qubit_name` is `True`
        each key `i` correspond to the `i-th` qubit in the graph.
        """
        final_layout = self.initial_layout = None
        for transpiler_pass in self.passes:
            if isinstance(transpiler_pass, Optimizer):
                transpiler_pass.connectivity = self.connectivity
                circuit = transpiler_pass(circuit)
            elif isinstance(transpiler_pass, Placer):
                transpiler_pass.connectivity = self.connectivity
                if self.initial_layout == None:
                    self.initial_layout = transpiler_pass(circuit)
                    final_layout = (
                        self.initial_layout
                    )  # This way the final layout will be the same as the initial layout if no router is used
                else:
                    raise_error(
                        TranspilerPipelineError,
                        "You are defining more than one placer pass.",
                    )
            elif isinstance(transpiler_pass, Router):
                transpiler_pass.connectivity = self.connectivity
                if self.initial_layout is not None:
                    circuit, final_layout = transpiler_pass(
                        circuit, self.initial_layout
                    )
                else:
                    raise_error(
                        TranspilerPipelineError, "Use a placement pass before routing."
                    )
            elif isinstance(transpiler_pass, Unroller):
                circuit = transpiler_pass(circuit)
            else:
                raise_error(
                    TranspilerPipelineError,
                    f"Unrecognised transpiler pass: {transpiler_pass}",
                )
        if self.int_qubit_names and final_layout is not None:
            final_layout = {int(key[1:]): value for key, value in final_layout.items()}
        return circuit, final_layout

    def is_satisfied(self, circuit: Circuit):
        """Returns ``True`` if the circuit respects the hardware connectivity and native gates, ``False`` otherwise.

        Args:
            circuit (:class:`qibo.models.circuit.Circuit`): circuit to be checked.

        Returns:
            (bool): satisfiability condition.
        """
        try:
            assert_connectivity(circuit=circuit, connectivity=self.connectivity)
            assert_decomposition(circuit=circuit, native_gates=self.native_gates)
            return True
        except ConnectivityError:
            return False
        except DecompositionError:
            return False

    def get_initial_layout(self):
        """Return initial qubit layout"""
        return self.initial_layout
